'''
    Dict utilities.

    Copyright 2013-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.

    To do

    Fix infinite recursion:
        File "site-packages/syr/dict.py", line 533 in resolve_instance
        File "site-packages/syr/dict.py", line 621 in resolve_obj
        File "site-packages/syr/dict.py", line 533 in resolve_instance
        File "site-packages/syr/dict.py", line 621 in resolve_obj
        File "site-packages/syr/dict.py", line 533 in resolve_instance
        File "site-packages/syr/dict.py", line 621 in resolve_obj
        ...

        Line 533 was 'd[name] = resolve_obj(attr(*[instance]))' and
        621 was 'value = resolve_instance(obj)'.

        In dictify.resolve_obj(), after check_circular_reference() we need
        to check if obj is None. Otherwise why call it?
        Maybe check_circular_reference() should return the earlier computed value.

        After
            elif isinstance(obj, dict):

                value = resolve_dict(obj, json_compatible=json_compatible)
                if debug: log('resolve_obj(%s) value is dict: %r' % (obj, value))

            if value is None:
        we can probably skip "if value is None:" and change the nested 'if's with 'elif's.

        If types.ModuleType is an instance, we need to check
        'isinstance(obj, types.ModuleType)' before 'is_class_instance(obj)'.
'''

import datetime
import json
import types
from collections.abc import Mapping

from denova.python.log import get_log
from denova.python.utils import last_exception, is_class_instance

log = get_log()

_unimplemented_types = set()
datetime_types = (datetime.timedelta, datetime.date, datetime.datetime, datetime.time, datetime.timezone)

class Simple(dict):
    ''' You can reference a field in a Simple object by either obj.name or
        obj['name']. It's ideal for going to and from json, or when you need to iterate through fields.

        Simple is both a class and a dict. It's also a very simple way to
        create a class instance.

        An alternative is the "dataclass" decorator.

        >>> simpl = Simple()

        >>> simpl.a = 1
        >>> simpl['b'] = 'hi'

        >>> simpl['a']
        1
        >>> simpl.b
        'hi'

        Or init with a dict, such as returned by json.dumps().

        >>> simpl = Simple({'x': 3, 'y': 'blue'})

        >>> simpl.x
        3
        >>> simpl['x']
        3
        >>> simpl.y
        'blue'
        >>> simpl['y']
        'blue'

        This class is very similar to javascript objects and django template dicts.

        No __init__ to define. Just pythonic duck typing with flexible
        access.

        It's especially useful when you often have to convert between
        class instances and dicts, such as with JSON.

        If you know some or all of the keys and values in advance, use
        Simple anywhere you would use {} or dict(). Code is must cleaner.
        Simple is also a great solution when you don't know all of the
        members of a class in advance. In that case init fields with just
        obj.name.

        You can initialize Simple with anything you can pass to dict().

        Keys and values which are of type dict are converted to Simple
        instances.

        To convert a class instance to a dict, see denova.python.dict.dictify().
        Python's builtin dict() does not convert inner instances of Simple,
        at least in python2.

        >>> d = {'a': 1, 'b': 'hi'}

        >>> simpl = Simple(d)
        >>> simpl.c = {'a': 1, 'b': 'hi'}

        >>> sorted(simpl.keys())
        ['a', 'b', 'c']

        >>> simpl.a
        1
        >>> simpl.b
        'hi'

        >>> sorted(simpl.c.keys())
        ['a', 'b']
        >>> simpl.c['a']
        1
        >>> simpl.c['b']
        'hi'
        >>> isinstance(simpl.c, Simple)
        True

        >>> simpl.d
        Traceback (most recent call last):
            ...
        AttributeError: d
        >>> 'd' in simpl
        False

        >>> simpl.d = 'hey'
        >>> 'd' in simpl
        True
        >>> simpl['d']
        'hey'

        >>> redict = dict(simpl)
        >>> type(redict) is dict
        True
        >>> sorted(redict.keys())
        ['a', 'b', 'c', 'd']

        >>> redict['a']
        1
        >>> redict['b']
        'hi'

        >>> isinstance(redict['c'], dict)
        True
        >>> sorted(redict['c'].keys())
        ['a', 'b']
        >>> redict['c']['a']
        1
        >>> redict['c']['b']
        'hi'

        >>> isinstance(redict['c'], Simple)
        True
        >>> redict['c'].a
        1
        >>> redict['c'].b
        'hi'

        >>> redict['d']
        'hey'

        >>> redict['e']
        Traceback (most recent call last):
            ...
        KeyError: 'e'

    '''

    def __getattr__(self, name):
        if name in self:
            value = self[name]
        else:
            raise AttributeError(name)
        return value

    def __setattr__(self, name, value):
        if isinstance(name, dict):
            name = Simple(name)
        if isinstance(value, dict):
            value = Simple(value)
        self[name] = value

    def __delattr__(self, name):
        if name in self.keys():
            del self[name]
        else:
            raise AttributeError(name)

    def __repr__(self):
        try:
            result = repr(dict(self))
        except (Exception, AttributeError, IOError, OSError):
            result = '>>>Simple repr error>>>'
        return result

    def __str__(self):
        try:
            result = str(dict(self))
        except (Exception, AttributeError, IOError, OSError):
            result = 'Simple str error'
        return result

    def __dir__(self):
        return self.keys()

    def __hash__(self):
        return id(self)

class CaseInsensitiveDict(Mapping):
    ''' Case insensitive dict.

        Dict lookups ignore key case. The key matches in lower, upper, or mixed case.

        Mostly from http://stackoverflow.com/questions/3296499/case-insensitive-dictionary-search-with-python
    '''

    def __init__(self, d=None):
        if d is None:
            d = {}
        self._d = d
        # _s is a dict mapping lowercase _d keys to actual _d keys
        self._s = dict((k.lower(), k) for k in d)

    def __contains__(self, k):
        return k.lower() in self._s

    def __len__(self):
        return len(self._s)

    def __iter__(self):
        return iter(self._s)

    def __getitem__(self, k):
        return self._d[self._s[k.lower()]]

    def __setitem__(self, k, v):
        self._d[k] = v
        self._s[k.lower()] = k

    def __delitem__(self, k):
        del self._d[self._s[k.lower()]]
        del self._s[k.lower()]

    def __str__(self):
        strings = []
        for key in self._d:
            strings.append(f'{key}: {self._d[key]}')
        return ', '.join(strings)

    def pop(self, k):
        k0 = self._s.pop(k.lower())
        return self._d.pop(k0)

    def actual_key_case(self, k):
        return self._s.get(k.lower())

def dictify(obj, deep=False, json_compatible=False, debug=False):
    ''' Resolves an object to a dictionary.

        If deep is True, recurses as needed. Default is False.

        Allowable object types are:
            NoneType,
            BooleanType,
            IntType, LongType, FloatType, ComplexType,
            StringTypes,
            TupleType, ListType, DictType,
            #MethodType, FunctionType,
            #GeneratorType,
            datetime.timedelta, datetime.date, datetime.datetime, datetime.time,
            class instances

        As of Python 2.6 2013-05-02 types.Instancetype is not reliable. We use
        denova.python.utils.is_class_instance().

        A class instance is converted to a dict, with only data members and
        functions without parameters. Functions that require parameters (beyond self)
        are ignored. For an object that accepts dot
        notation so you may not have to change your code, see
        denova.python.dict.Simple(). Builtin instance member names beginning
        with '__' are ignored.

        dictify tries to use the current return value from methods.
        The method is called with no args (except self, i.e. the instance).
        We try a static method with no arguments. If these fail,
        the method is ignored.

        dictify converts a generator to a list. Warning: This may result in
        large, or even infinite, lists.

        json_compatible makes dictionary keys compatible with json, i.e. one of
        (str, unicode, int, long, float, bool, None).

        To do:
            dictify() could also find why json.dumps() doesn't always work.
            Add a param 'test_json=True'. When set call json.dumps() on
            every object.Log or raise exception on errors. Probabaly
            the deepest/last error is most important.

        >>> from denova.python.format import pretty

        >>> class Test(object):
        ...     a = 1
        ...     b = 'hi'
        ...     def __init__(self):
        ...         self.c = 2
        ...         self.d = 'hey'
        ...         self.e = datetime.timedelta(weeks=1, days=2, hours=3,
        ...                                     minutes=4, seconds=5)
        ...         self.f = datetime.date(2000, 1, 2)
        ...         self.g = datetime.datetime(2000, 1, 2, 3, 4, 5, 6)
        ...         self.h = datetime.time(1, 2, 3, 4)

        >>> test = Test()
        >>> print(pretty(dictify(test)))
        {
            'a': 1,
            'b': 'hi',
            'c': 2,
            'd': 'hey',
            'e': {
                'days': 9,
                'microseconds': 0,
                'seconds': 11045,
            },
            'f': {
                'day': 2,
                'month': 1,
                'year': 2000,
            },
            'g': {
                'day': 2,
                'hour': 3,
                'microsecond': 6,
                'minute': 4,
                'month': 1,
                'second': 5,
                'tzinfo': None,
                'year': 2000,
            },
            'h': {
                'hour': 1,
                'microsecond': 4,
                'minute': 2,
                'second': 3,
                'tzinfo': None,
            },
        }

        >>> import datetime
        >>> dt = datetime.date(2001, 12, 1)
        >>> dictified = dictify(dt)
        >>> sorted(dictified.keys())
        ['day', 'month', 'year']
        >>> dictified['year']
        2001
        >>> dictified['month']
        12
        >>> dictified['day']
        1

        >>> dt = {1: datetime.date(2002, 12, 1)}
        >>> print(pretty(dictify(dt)))
        {
            1: {
                'day': 1,
                'month': 12,
                'year': 2002,
            },
        }

        >>> class OldStyleClass:
        ...     class_data = 27
        ...
        ...     def __init__(self):
        ...         self.instance_data = 'idata'

        ...     def instance_function(self):
        ...         return 3
        >>> old_c = OldStyleClass()
        >>> print(pretty(dictify(old_c)))
        {
            'class_data': 27,
            'instance_data': 'idata',
        }

        >>> class NewStyleClass(object):
        ...     class_data = 27
        ...
        ...     def __init__(self):
        ...         self.instance_data = 'idata'

        ...     def instance_function(self):
        ...         return 3
        >>> new_c = NewStyleClass()
        >>> print(pretty(dictify(new_c)))
        {
            'class_data': 27,
            'instance_data': 'idata',
        }

        >>> from datetime import timezone
        >>> dictify(timezone.utc)

        >>> try:
        ...     import pytz
        ...     dictify(pytz.timezone('UTC').utcoffset(None))
        ... except ImportError:
        ...     pass
        {'days': 0, 'seconds': 0, 'microseconds': 0}
        >>> dictify({'_dst': {'days': 0, 'microseconds': 0, 'seconds': 0}, '_tzname': 'UTC', '_utcoffset': {'days': 0, 'microseconds': 0, 'seconds': 0}, 'zone': 'UTC'})
        {'_dst': {'days': 0, 'microseconds': 0, 'seconds': 0}, '_tzname': 'UTC', '_utcoffset': {'days': 0, 'microseconds': 0, 'seconds': 0}, 'zone': 'UTC'}

        # if the following line does not have the new line escaped, then python rejects the formating of the dict
        >>> timing_access_line = {'line': '207.241.233.238 - - [16/Oct/2019:19:11:06 +0000] "HEAD /page_timing/?nt_red_cnt=1&nt_nav_type=0&nt_nav_st=1571253056365&nt_red_st=1571253056365&nt_red_end=1571253057865&nt_fet_st=1571253057865&nt_dns_st=1571253057865&nt_dns_end=1571253057865&nt_con_st=1571253057865&nt_con_end=1571253057865&nt_req_st=1571253057866&nt_res_st=1571253060162&nt_res_end=1571253060349&nt_domloading=1571253060345&nt_domint=1571253060738&nt_domcontloaded_st=1571253060738&nt_domcontloaded_end=1571253060747&nt_domcomp=1571253065139&nt_load_st=1571253065140&nt_load_end=1571253065146&nt_unload_st=0&nt_unload_end=0&u=http%3A%2F%2Fweb.archive.org%2Fweb%2F20160128095330%2Fhttp%3A%2F%2Fwww.denova.com%2F&v=0.9&vis.st=visible HTTP/1.1" 301 0 "-" "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0.2 Safari/605.1.15"\\n', 'ip_address': b'207.241.233.238', 'domain': '', 'method': b'HEAD', 'timestamp': datetime.datetime(2019, 10, 16, 19, 11, 6, tzinfo=datetime.timezone.utc), 'referer': '', 'url': b'/page_timing/', 'query_string': b'nt_red_cnt=1&nt_nav_type=0&nt_nav_st=1571253056365&nt_red_st=1571253056365&nt_red_end=1571253057865&nt_fet_st=1571253057865&nt_dns_st=1571253057865&nt_dns_end=1571253057865&nt_con_st=1571253057865&nt_con_end=1571253057865&nt_req_st=1571253057866&nt_res_st=1571253060162&nt_res_end=1571253060349&nt_domloading=1571253060345&nt_domint=1571253060738&nt_domcontloaded_st=1571253060738&nt_domcontloaded_end=1571253060747&nt_domcomp=1571253065139&nt_load_st=1571253065140&nt_load_end=1571253065146&nt_unload_st=0&nt_unload_end=0&u=http%3A%2F%2Fweb.archive.org%2Fweb%2F20160128095330%2Fhttp%3A%2F%2Fwww.denova.com%2F&v=0.9&vis.st=visible', 'http_status_code': 301, 'protocol': b'HTTP/1.1', 'bytes': 0, 'user': '', 'agent': b'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0.2 Safari/605.1.15', 'browser_name': 'Mozilla', 'browser_version': '5.0', 'other': '(Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0.2 Safari/605.1.15', 'trace_url_found': False}
        >>> d = dictify(timing_access_line)
        >>> isinstance(d, Simple)
        True
    '''

    allowed_types = (
            type(None),
            bool,
            int, int, float, complex,
            # ?? StringTypes is itself a tuple; can we nest it like this?
            str,
            tuple, list, dict,
            object,
            # types.MethodType, types.FunctionType,
            # types.ModuleType,
            # types.GeneratorType,
            ) + datetime_types

    def type_allowed(obj):
        return isinstance(obj, allowed_types)
        #return (isinstance(obj, allowed_types) or
        #    is_class_instance(obj))

    def check_circular_reference(obj):
        ''' Check for circular references to instances. '''

        # do not check classes that we resolve specially
        #if is_class_instance(obj) and not isinstance(obj, datetime_types):
        if not isinstance(obj, datetime_types):

            obj_id = id(obj)
            if obj_id in obj_ids:
                obj = f'__syr.dict.dictify: circular_reference to {repr(obj)}, type {type(obj)}__'
                if debug: log.warning(obj)

            else:
                obj_ids.add(obj_id)

        return obj

    def resolve_string(obj):
        ''' Return object if unicode, else return str.

            Starting with python 2.6 all strings are unicode.
            But for readability we use a plain string where possible. '''

        try:
            value = str(obj)
        except (Exception, AttributeError, IOError, OSError):
            value = f'{obj}'

        return value

    def resolve_datetime(obj):
        if isinstance(obj, datetime.timedelta):
            value = Simple({
                'days': obj.days,
                'seconds': obj.seconds,
                'microseconds': obj.microseconds,
                })
        elif isinstance(obj, datetime.datetime):
            value = Simple({
                'year': obj.year,
                'month': obj.month,
                'day': obj.day,
                'hour': obj.hour,
                'minute': obj.minute,
                'second': obj.second,
                'microsecond': obj.microsecond,
                'tzinfo': obj.tzinfo,
                })
        elif isinstance(obj, datetime.date):
            value = Simple({
                'year': obj.year,
                'month': obj.month,
                'day': obj.day,
                })
        elif isinstance(obj, datetime.time):
            value = Simple({
                'hour': obj.hour,
                'minute': obj.minute,
                'second': obj.second,
                'microsecond': obj.microsecond,
                'tzinfo': obj.tzinfo,
                })
        else:
            value = None

        if debug: log(f'datetime obj: {obj!r}, type: {type(obj)}, value: {value!r}')
        return value

    def resolve_dict(obj, json_compatible=False):
        ''' Resolve a simple dict object. This is a deep resolve to a Simple. '''

        d = Simple({})
        for key in obj:

            # a dictionary needs a hashable key
            try:
                new_key = resolve_obj(key)
                hash(new_key)
            except (Exception, AttributeError, IOError, OSError):
                new_key = f'{type(key)}-{id(key)}'

            if json_compatible:
                # convert key to a json compatible type
                if not isinstance(new_key, (str, int, float, bool, None)):
                    new_key = repr(new_key)

            new_value = resolve_obj(obj[key])

            try:
                d[new_key] = new_value
            except TypeError: # e.g. unhashable type
                if debug:
                    log.warning('resolving Simple, got TypeError')
                    log(f'    key is {key}, type {type(key)}')
                    log(f'    new_key is {new_key}, type {type(new_key)}')
                    log(f'    obj[key] is {obj[key]}, type {type(obj[key])}')
                    log(f'    new_value is {new_value}, type {type(new_value)}')
                    log(last_exception())
                d[key] = new_value

        return d

    def resolve_instance(obj):
        ''' Resolve an instance of a class. '''

        d = Simple({})

        # call the obj an instance for clarity here
        instance = obj
        # get names of instance attributes
        for name in dir(instance):
            # ignore builtins, etc.
            if not name.startswith('__'):

                # usually commented out - set debug to test special case
                #if name == 'previous_conversion': #DEBUG
                #    _old_debug = debug #DEBUG
                #    debug = True #DEBUG

                if debug: log('in resolve_obj() getting "%s" attribute "%s"' %
                    (repr(instance), name))

                try:
                    attr = getattr(instance, name)
                    if debug: log(
                        'in resolve_obj() instance: "%s", attribute: "%s", type: %s' %
                        (repr(instance), name, type(attr)))

                    # convert name to an allowed type
                    # infinite recursion
                    # name = resolve_obj(name)

                    if type_allowed(attr):

                        # if this attr is a method object
                        if isinstance(attr, types.MethodType):
                            if debug: log(f'{attr!r} is a MethodType, so skipping')

                        # if this attr is a static method object
                        elif isinstance(attr, types.FunctionType) and deep:
                            if debug: log(f'{attr!r} is a FunctionType, so skipping')

                        else:
                            if debug: log('member {}.{} is an allowed type ({}) so resolving object'.
                                              format(name, repr(attr), type(attr)))
                            d[name] = resolve_obj(attr)

                    else:
                        if debug: log.warning(f'in resolve_obj() type not allowed ({attr!r}), so skipping')

                except (Exception, AttributeError, IOError, OSError):
                    # these seem to be caused by using @property, making it hard to
                    # get a function attr without calling the function
                    if debug:
                        log('in resolve_obj() ignoring following exception')
                        log(last_exception())

                # usually commented out - set debug to test special case
                #if name == 'previous_conversion': #DEBUG
                #    debug = _old_debug #DEBUG

        return d

    def resolve_module(module):
        ''' Resolve a module to a dict object. '''

        d = Simple({})
        for k, v in list(module.__dict__.items()):
            if not k.startswith('__'):
                d[k] = v

        return d

    def resolve_obj(obj):
        ''' Resolve any type to a dict object. '''

        #assert not debug #DEBUG
        # usually commented out - set debug locally to test special case
        # if isinstance(obj, datetime_types): #DEBUG
        #     debug = True #DEBUG

        value = None

        if debug: log(f'resolve_obj({obj}) type {type(obj)}')
        obj = check_circular_reference(obj)

        if isinstance(obj, str):
            value = resolve_string(obj)

        elif isinstance(obj, (tuple, types.GeneratorType)):

            # immutable iterators
            value = tuple(resolve_obj(item) for item in obj)
            if debug: log(f'resolve_obj({obj}) value is tuple: {value!r}')

        elif isinstance(obj, list):

            # mutable iterator
            value = list(resolve_obj(item) for item in obj)
            if debug: log(f'resolve_obj({obj}) value is list: {value!r}')

        elif isinstance(obj, dict):

            value = resolve_dict(obj, json_compatible=json_compatible)
            if debug: log(f'resolve_obj({obj}) value is dict: {value!r}')

        elif isinstance(obj, datetime_types):

            value = resolve_datetime(obj)

        elif isinstance(obj, datetime.tzinfo):

            # obviously wrong for some systems, but not ours
            value = None

        elif is_class_instance(obj):

            value = resolve_instance(obj)
            if debug: log(f'resolve_obj({obj}) value is instance: {value!r}')

        elif isinstance(obj, types.ModuleType):

            value = resolve_module(obj)
            if debug: log(f'resolve_obj({obj}) value is module: {value!r}')

        elif type_allowed(obj):

            value = obj
            if debug: log(f'resolve_obj({obj}) value is allowed type: {type(obj)}')

        else:

            # any other type as just the type, not restorable
            value = str(type(obj))
            if value not in _unimplemented_types:
                # mention each type just once
                _unimplemented_types.add(value)
                if debug: log.warning(f'resolve_obj({obj}) value is unimplemented type: {value}')

        if debug: log(f'resolve_obj({obj}) final type: {type(value)}, value: {value!r}')
        assert not isinstance(value, datetime_types) #DEBUG

        # remove the object from circular reference check
        if id(obj) in obj_ids:
            obj_ids.remove(id(obj))

        return value

    debug = debug

    if debug: log(f'dictify({obj})')
    obj_ids = set()

    value = resolve_obj(obj)

    # if debug: log('dictify(%s) is %r' % (object_name(obj, include_repr=True), value))
    return value

def force_json(obj, debug=False):
    ''' Return json for the object, dropping fields as needed. '''

    try:
        json_out = json.dumps(obj)
    except TypeError as t_error:
        t_error_str = str(t_error)
        log(t_error_str)
        if 'is not JSON serializable' in t_error_str:

            log(f'json.dumps({type(obj)}) failed; try dictify()')
            d = dictify(obj, debug=debug)
            json_out = json.dumps(d)

        else:
            # we don't know what happened, so raise the exception
            raise

    return json_out


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Custom django template tags and filters

    Copyright 2010-2020 DeNova
    Last modified: 2020-10-25

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import urllib.request

try:
    from django.conf import settings
    import django.forms
    from django.template import Library, Node
    import django.templatetags.static
    import django.utils.html
    from django.utils.safestring import mark_safe
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.django_addons.templatetags.data_img import data_img
from denova.django_addons.templatetags.lookup import lookup
from denova.django_addons.templatetags.var import do_variables
from denova.django_addons.utils import get_absolute_url
from denova.net.html_addons import strip_whitespace_in_html
from denova.python.log import get_log
from denova.python.times import timedelta_to_human_readable

log = get_log()

register = Library()

static_url_prefix = get_absolute_url(settings.STATIC_URL, settings.CONTENT_HOME_URL)

# register imported tags and filters that we want to include when we load 'custom'
# re-registering django.templatetags.static.static apparently works
register.simple_tag(django.templatetags.static.static)
register.filter('data_img', data_img)
register.filter('lookup', lookup)
register.tag('var', do_variables)

@register.simple_tag
def get_title(title):
    # this is a special case
    title = settings.TOP_LEVEL_DOMAIN.title()
    return title

@register.simple_tag
def transcode(url):

    '''
    Template tag to get the contents of a url.

    You can't transcode local urls because django is not re-entrant.
    Instead, in the view set a template variable with the local content.

    This code assumes a url starting with STATIC_URL is served from a different server,
    and so is ok.

    Example usage:

        {% transcode 'http://example.com' %}

    Returns:

        ... contents of http://example.com page ...

    '''

    def url_ok(url):
        ''' A url is ok if it is static or not on the local server '''

        return (
            url.startswith(static_url_prefix) or
            not url.startswith(settings.CONTENT_HOME_URL))

    url = url.strip()
    url = get_absolute_url(url, settings.CONTENT_HOME_URL)
    assert url_ok(url), ValueError('transcode can only include content from other servers')
    stream = urllib.request.urlopen(url)
    contents = stream.read()
    stream.close()
    return contents


class MinWhiteSpaceNode(Node):

    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        ''' Collect the strings from all nodes into a single string, and
            remove excess white space from the entire string.'''

        mintext = True
        return strip_whitespace_in_html(self.nodelist.render(context), mintext)

def minwhitespace(parser, token):
    '''
        Template tag to remove excess white space.

        The Django builtin tag "spaceless" removes whitespace between html tags.
        This tries to remove as much white space as possible without changing the
        appearance of the rendered html.

        Warning: removes repeated whitespace everywhere, such as in embedded
        javascript and between <pre> html tags.

        Example usage:

            {% minwhitespace %}
                <p>
                    <a href="foo/">Foo</a>

                    Bar

                </p>

            {% endminwhitespace %}

        Returns:

            <p><a href="foo/">Foo</a>Bar</p>

    '''
    nodelist = parser.parse(('endminwhitespace',))
    parser.delete_first_token()
    return MinWhiteSpaceNode(nodelist)


class StripWhitespaceNode(Node):

    def __init__(self, nodelist):
        self.nodelist = nodelist

    def render(self, context):
        ''' Collect the strings from all nodes into a single string, and
            remove leading and trailing white space from the entire string.'''

        return self.nodelist.render(context).strip()

@register.simple_tag
def stripwhitespace(parser, token):
    '''
    Template tag to remove leading and trailing white space.

    Example usage::

        {% stripwhitespace %}




            <p>
                <a href="foo/">Foo</a>
                Bar
            </p>



        {% endstripwhitespace %}

    Returns::
        <p>
            <a href="foo/">Foo</a>
            Bar
        </p>

    '''
    nodelist = parser.parse(('endstripwhitespace',))
    parser.delete_first_token()
    return StripWhitespaceNode(nodelist)

def strip(value):
    ''' Strip text filter. Uses string.strip(). '''
    return mark_safe(value.strip())

def timedelta_in_words(value):
    ''' Timedelta to human readable template filter.'''
    return mark_safe(timedelta_to_human_readable(value))

def blankline():
    ''' Template tag for a bootstrap css blank row.

        Example usage:

            {% blankline %}

        Returns:

            <div class="block"> &nbsp; </div>
    '''

    return mark_safe('<div class="container block"> &nbsp; </div>')

def bullet():
    ''' Template tag for an inline bullet point separator.

        In HTML:
        char   description          unicode   html       html entity    utf-8

        ·      Middle Dot           U+00B7    &#183;     &middot;       C2 B7
        •      Bullet               U+2022    &#8226;    &bull;         E2 80 A2
‧       -      Hyphenation Point    U+2027    &#8321;                   E2 80 A7
        ∙      Bullet Operator      U+2219    &#8729;                   E2 88 99
        ●      Black Circle         U+25CF    &#9679;                   E2 97 8F
        ⬤      Black Large Circle   U+2B24    &#11044;

        In CSS:
            \00B7

        Example usage:

            {% bullet %}

        Returns:

            (see html below)

    '''

    html = '&bull;'

    return mark_safe(html)


register.tag(minwhitespace)
register.filter(timedelta_in_words)
register.filter(strip)
register.simple_tag(blankline)
register.simple_tag(bullet)

@register.simple_tag
def title(*args):
    ''' Title text.

        Usage::

            {% title "My title" %}

        Use 'title'/'endtitle' if there are django template variables in the text
        or the text is more than one line.
    '''

    if args:
        text = args[0]
        html = django.utils.html.format_html('<h3> <strong> {} </strong> </h3>', text)
    else:
        html = django.utils.html.format_html('<h3> <strong> ')
    return html

@register.simple_tag
def endtitle():
    ''' End variable or multiline title text.

        Usage::

            {% title %}
                My title
            {% endtitle %}
    '''

    return django.utils.html.format_html('</strong> </h3>')

@register.simple_tag
def subtitle(*args):
    ''' NYT style subtitle.

        Usage::

            {% subtitle "My subtitle" %}

        Use 'subtitle'/'endsubtitle' if there are django template variables in the text
        or the text is more than one line.
    '''

    if args:
        text = args[0]
        html = django.utils.html.format_html('<h4> <strong> {} </strong> </h4>', text)
    else:
        html = django.utils.html.format_html('<strong> <h4>')
    return html

@register.simple_tag
def endsubtitle():
    ''' End variable or multiline NYT style subtitle.

        Usage::

            {% subtitle %}
                My subtitle
            {% endsubtitle %}
    '''

    return django.utils.html.format_html(' </strong> </h4>')

@register.simple_tag
def link_button(href, label,
                target=None, tooltip=None,
                onclick=None, active=True,
                primary=False, btnblock=False):
    '''
        Link shown as a button with consistent style.

        This button can be used anywhere a link is allowed.

        If your label includes template variables or filters,
        use the start_link_button/end_link_button form.

        To reference the button in javascript, make the "label" all
        lower case and replace spaces with dashes. Use the result
        as the button's name by suffixing "-button" and its id by
        prefixing "id-".

        Usage: {% link_button "My href" "My label" %}
    '''

    html = start_link_button(href, label,
                                   target=target, tooltip=tooltip,
                                   onclick=onclick, active=active,
                                   primary=primary, btnblock=btnblock)
    html = html + django.utils.html.format_html(label)
    html = html + end_link_button()

    return html

@register.simple_tag
def start_link_button(href,
                      label=None,
                      target=None, tooltip=None,
                      onclick=None, active=True,
                      primary=False, btnblock=False):
    '''
        Link shown as a button with consistent style.

        This button can be used anywhere a link is allowed.

        Usage:
            {% start_link_button "My href" "Invisible label" %}

                My visible label

            {% end_link_button %}

        There are actually 2 labels for this button.  The invisible label
        in the tag is used only to create id= and name=. The visible label
        between start_link_button and end_link_button is the one users see.
        This allows you to change the visible label without changing tests
        or javascript that reference the tag id or name.

        To reference the button in javascript, make the "label" all
        lower case and replace spaces with dashes. Use the result
        as the button's name by suffixing "-button" and its id by
        prefixing "id-".

    '''

    if primary:
        button_type = 'btn-primary'
    else:
        button_type = 'btn-secondary'

    if active:
        active_class = ''
    else:
        active_class = ' disabled'

    # make the buttons fill the block; good when multiple buttons in table
    if btnblock:
        block = ' btn-block'
    else:
        block = ''

    name, id_tag = button_name_id(label)

    extra = ''
    if target:
        extra += f' target="{target}"'
    if tooltip:
        extra += f' title="{tooltip}"'
    if onclick:
        extra += f' onclick="{onclick}"'

    BaseFormat = '<a href="{}" name="{}" id="{}" class="btn {}{}{}" role="button" {}> <strong>'
    html = django.utils.html.format_html(BaseFormat.format(href, name, id_tag, button_type, active_class, block, extra))

    return html

@register.simple_tag
def end_link_button():
    ''' End variable or multiline NYT style subtitle.

        Usage::

            {% start_link_button %}
                My label
            {% endsubtitle %}
    '''

    return django.utils.html.format_html(' </strong> </a>')

@register.simple_tag
def form_button(value, tooltip=None, name=None, id_tag=None, onclick=None, primary=True, btnblock=False, btntype=None):
    '''
        Consistent button tag style.

        This button is used in a form.

        The element's name is "VALUE-button". The element's id is "VALUE-id".
        VALUE is the button text value capitalized.

        To reference the button in javascript, make the "label" all
        lower case and replace spaces with dashes. Use the result
        as the button's name by suffixing "-button" and its id by
        prefixing "id-".

        Usage: {% form_button "Value" %}
    '''

    BaseFormat = '<input type="{}" value="{}" name="{}" id="{}" alt="{}" class="btn {} font-weight-bold {}" role="button" {}/>'

    if name is None and id_tag is None:
        name, id_tag = button_name_id(value)
    elif name is None:
        name, __ = button_name_id(value)
    elif id_tag is None:
        __, id_tag = button_name_id(value)

    if not btntype:
        btntype = 'button'

    extra = ''
    if tooltip:
        extra += f' title="{tooltip}"'
    if onclick:
        extra += f' onclick="{onclick}"'

    if primary:
        button_type = 'btn-primary'
    else:
        button_type = 'btn-secondary'

    # make the buttons fill the block; good when multiple buttons in table
    if btnblock:
        block = ' btn-block'
    else:
        block = ''

    btn_html = BaseFormat.format(btntype, value, name, id_tag, value, button_type, block, extra)
    html = django.utils.html.format_html(btn_html)

    return html


@register.simple_tag
def submit_button(*args, **kwargs):
    '''
        Consistent submit button style.

        This button is used in a form to submit form data.

        See button() for parameters.

        Usage: {% submit_button "Value" %}
    '''

    return form_button(*args, btntype='submit', **kwargs)


@register.simple_tag
def reset_button(*args, **kwargs):
    '''
        Consistent reset button style.

        This button is used in a form to reset form data.

        See button() for parameters.

        Usage: {% submit_button "Value" %}
    '''

    return form_button(*args, btntype='reset', **kwargs)


@register.simple_tag
def card(*args):
    ''' Our standard version of the bootstrap card.

        Usage::

            {% card "My card" %}

        Use 'card'/'endcard' if there are django template variables in the content
        or the content is more than one line.
    '''

    START_CARD_HTML = '<div class="card"><div class="card-body border rounded" style="background-color:whitesmoke">'

    if args:
        content = args[0]
        html = django.utils.html.format_html('{} {} {}', START_CARD_HTML, content, endcard())
    else:
        html = django.utils.html.format_html(START_CARD_HTML)
    return html

@register.simple_tag
def endcard():
    ''' End variable or multiline card content.

        Usage::

            {% card %}
                My card
            {% endcard %}
    '''

    return django.utils.html.format_html('</div> </div>')

@register.simple_tag
def form_as_table(form):
    ''' Render form as table.

        Django version is junk.

        To do: allow for "{% trans 'Email' %}:" labels.
    '''

    html = '<table>'
    for field in form.visible_fields():

        html = html + '<tr>'

        html = html + '<th>'
        html = html + field.label + ':'
        html = html + '</th>'

        html = html + '<td>'
        html = html + str(field)
        if field.errors:
            html = html + ('<br/>' +
                           '<font color="red">' +
                           str(field.errors) +
                           '</font>' +
                           '<br/>')
        html = html + '</td>'

        html = html + '</tr>'

    html = html + '</table>'

    return django.utils.html.format_html(html)

@register.simple_tag
def form_as_narrow(form):
    ''' Render narrow form. '''

    html = ''
    for field in form.visible_fields():

        html = html + field.label + ':' + '<br/>'
        html = html + '&nbsp;&nbsp;&nbsp;&nbsp;' + str(field) + '<br/>'
        if field.errors:
            html = html + ('&nbsp;&nbsp;&nbsp;&nbsp;' +
                           '<font color="red">' +
                           str(field.errors) +
                           '</font>' +
                           '<br/>')

    return django.utils.html.format_html(html)

def button_name_id(text):
    ''' Return button name and id from button label. '''

    base = text.replace(' ', '-').replace('.', '').replace(',', '').replace("'", '').replace("?", '').replace("<br/>", '-').replace('&nbsp;', '').lower()
    name = f'{base}-button'
    id_tag = f'{base}-id'

    return name, id_tag



'''
    Python programming utilities.

    Utility classes and functions. Mostly utiities about python,
    not just in python.

    This module also holds code that hasn't been categorized into
    other packages. For example, many functions could go in
    denova.os.fs.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import bz2
import gzip as gz
import importlib
import io
import locale
import os
import os.path
import string
import re
import sys
import trace
import traceback
import types
import unicodedata
import zipfile
from contextlib import contextmanager
from fnmatch import fnmatch
from glob import glob
from io import StringIO
from subprocess import CalledProcessError
from urllib.parse import urlparse
from tempfile import gettempdir

#from denova.python._log import log

version = sys.version_info[0]
prefered_encoding = 'utf-8'

# we can't use denova.python.log here because denova.python.log uses this module
_debug = False
if _debug:
    def log(msg):
        print(msg)
else:
    def log(msg):
        pass

Http_Separator = '://'

# linux allows almost any characters, but this is cross platform pathnames
valid_pathname_chars = f"-_.()/\\: {string.ascii_letters}{string.digits}"

class NotImplementedException(Exception):
    ''' Operation not implemented exception. '''
    pass

class MovedPermanentlyException(Exception):
    ''' Object moved permanently exception.

        Always say where it was moved. '''
    pass

def set_default_encoding(encoding=prefered_encoding):
    '''
        >>> set_default_encoding(prefered_encoding)
        >>> encoding = sys.getdefaultencoding()
        >>> encoding == 'utf-8'
        True
    '''

    # only available in python2
    pass

def dynamically_import_module(name):
    '''
        Dynamically import a module. See python docs on __import__()

        >>> module_str = str(dynamically_import_module('denova.python'))
        >>> module_str.startswith("<module 'denova.python' from ")
        True
        >>> module_str.endswith("denova/python/__init__.py'>")
        True
     '''

    module = __import__(name)
    components = name.split('.')
    for component in components[1:]:
        module = getattr(module, component)
    return module

def dynamic_import(name):
    '''
        >>> module_str = str(dynamic_import('denova.python'))
        >>> module_str.startswith("<module 'denova.python' from ")
        True
        >>> module_str.endswith("denova/python/__init__.py'>")
        True
    '''
    # from Python Library Reference, Built-in Functions, __import__
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def print_imported_modules(filename):
    ''' Print modules imported by filename.

        Warning: Runs filename.

        >>> # how do we test this? how we generally use it:
        >>> # print_imported_modules(__file__)
    '''

    import modulefinder

    if filename.endswith('.pyc') or filename.endswith('.pyo'):
        filename = filename[:-1]
    print(filename)

    finder = modulefinder.ModuleFinder()
    finder.run_script(filename)

    # See /usr/share/doc/python2.7/html/library/modulefinder.html
    print('Loaded modules:')
    for name in sorted(finder.modules.keys()):
        mod = finder.modules[name]
        print(f'    {name}: {mod}')
        globalnames = list(mod.globalnames.keys())
        modules = sorted(globalnames[:3])
        print(','.join(modules))

    print('Modules not imported:')
    keys = finder.badmodules.keys()
    for name in keys:
        print(f'    {name}')

def object_name(obj, include_repr=False):
    ''' Get a human readable type name of a module, function, class, class method, or instance.

        The name is not guaranteed to be unique.

        If include_repr is True, an instance has its string representation appended.

        >>> import datetime
        >>> object_name(datetime.date)
        'datetime.date'

        >>> object_name('test string')
        'builtins.str instance'

        >>> class TestClass():
        ...     pass
        >>> name = object_name(TestClass)
        >>> name == 'denova.python.utils.TestClass' or name == 'utils.TestClass'
        True
        >>> inst = object_name(TestClass())
        >>> inst == 'denova.python.utils.TestClass instance' or inst == 'utils.TestClass instance'
        True

        To do: Consider using 'inspect' module.
    '''

    #package_name = getattr(obj, '__package__', None)
    module_name = getattr(obj, '__module__', None)
    local_object_name = getattr(obj, '__name__', None)

    """
    print(f'obj: {obj}, type(obj): {type(obj)}') #DEBUG
    print(f'initial package_name: {package_name}, type(package_name): {type(package_name)}') #DEBUG
    print(f'initial module_name: {module_name}, type(module_name): {type(module_name)}') #DEBUG
    print(f'initial local_object_name: {local_object_name}, type(local_object_name): {type(local_object_name)}') #DEBUG
    """

    name = None

    if module_name and local_object_name:
        name = f'{module_name}.{local_object_name}'

    elif local_object_name:
        name = local_object_name

    else:
        class_obj = getattr(obj, '__class__', None)
        class_module_name = getattr(class_obj, '__module__', None)
        local_object_name = getattr(class_obj, '__name__', None)

        """
        print(f'obj.__class__: {obj.__class__}, type(obj.__class__): {type(obj.__class__)}') #DEBUG
        print(f'class_obj: {class_obj}, type(class_obj): {type(class_obj)}') #DEBUG
        print(f'class_module_name: {class_module_name}, type(class_module_name): {type(class_module_name)}') #DEBUG
        print(f'local_object_name: {local_object_name}, type(local_object_name): {type(local_object_name)}') #DEBUG
        """

        name = f'{class_module_name}.{local_object_name} instance'
        if include_repr:
            name = f'{name}: {repr(obj)}'

    return name

def caller_id(ignore=None):
    ''' Return standard caller id string. '''

    filename, line = caller(ignore=ignore)
    return f'{filename}:{line}'

def caller(ignore=None, this_module_valid=False):
    ''' Returns (filename, linenumber) of the caller.

        To ignore calls from the current module::
            filename, line = caller(ignore=[__file__])
        This could conceivably ignore extra files if __file__ contains '.' etc.

        If this function is called from this module, set this_module_valid=True.
    '''

    def ignore_filename(filename):
        ''' Ignore files in ignore list and runpy.py. '''

        ignored = False
        if filename.endswith('/runpy.py'):
            ignored = True
        else:
            for pattern in _ignore:
                if re.match(pattern, filename):
                    ignored = True

        return ignored

    if ignore:
        _ignore = ignore
    else:
        _ignore = []
    # ignore denova.python.utils unless this_module_valid=True
    if not this_module_valid:
        _ignore = _ignore + [__file__]

    call_line = None

    stack = list(traceback.extract_stack())
    stack.reverse()
    for filename, linenumber, _, _ in stack:
        if not call_line:
            if not ignore_filename(filename):
                __, __, filename = filename.rpartition('/') # DEBUG
                call_line = (filename, linenumber)

    return call_line

def caller_module_name(ignore=None, this_module_valid=False):
    ''' Get the caller's fully qualified module name.

        If this function is called from this module, set this_module_valid=True.

        To do: Test linked package dirs in parent dirs.

        To get the parent caller instead of the module that actually
        calls caller_module_name():

            name = caller_module_name(ignore=[__file__])

        >>> # this code really needs to be tested from a different module
        >>> name = caller_module_name(this_module_valid=True)
        >>> name.endswith('utils')
        True
        >>> name = caller_module_name(ignore=[__file__], this_module_valid=True)
        >>> 'caller_module_name' in name
        True
    '''

    def ignore_filename(filename):
        ''' Ignore files in ignore list and runpy.py. '''

        if _debug_caller_module_name: print(f'in ignore_filename() ignore: {repr(ignore)}') #DEBUG
        return (filename in ignore) or filename.endswith('/runpy.py')

    def is_python_module_dir(dirname):
        # a python module dir has an __init__.py file
        init_path = os.path.join(dirname, '__init__.py')
        return os.path.exists(init_path)

    def strip_py(filename):
        if filename.endswith('.py'):
            filename, _, _ = filename.rpartition('.')
        return filename

    _debug_caller_module_name = False

    if _debug_caller_module_name: print(f'ignore: {repr(ignore)}') #DEBUG

    # make ignore list
    if ignore:
        _ignore = ignore
    else:
        _ignore = []
    # ignore denova.python.utils unless this_module_valid=True
    if not this_module_valid:
        _ignore = _ignore + [__file__]
    # make stack and __file__ filenames match
    ignore = []
    for filename in _ignore:
        # the ignore=[...] param is usually ignore=[__file__]
        # stack filenames end in .py; __file__ filenames end in .pyc or .pyo
        # make them all .py
        basename, _, extension = filename.rpartition('.')
        if extension == 'pyc' or extension == 'pyo':
            filename = basename + '.py'
        ignore.append(filename)

    name = None

    if _debug_caller_module_name:
        print('caller_module_name traceback.extract_stack():') #DEBUG
        for stack_item in traceback.extract_stack():
            print(f'    {stack_item}') #DEBUG

    stack = list(traceback.extract_stack())
    stack.reverse()
    # filename, line number, function name, text
    for filename, _, _, _ in stack:

        if not name:
            if ignore_filename(filename):
                if _debug_caller_module_name: print(f'caller_module_name ignored filename: {filename}') #DEBUG

            else:
                if _debug_caller_module_name: print(f'caller_module_name filename: {filename}') #DEBUG
                # find python module dirs in filename
                modules = []
                dirname, _, basename = filename.rpartition('/')
                while dirname and is_python_module_dir(dirname):
                    #if _debug_caller_module_name: print('caller_module_name is_python_module_dir: {}'.format(dirname)) #DEBUG
                    modules.append(os.path.basename(dirname))
                    dirname, _, _ = dirname.rpartition('/')
                modules.reverse()

                # if the filename is a __main__.py for a package, just use the package name
                if basename != '__main__.py':
                    modules.append(strip_py(basename))
                #if _debug_caller_module_name: print('caller_module_name modules: {}'.format(repr(modules))) #DEBUG

                name = '.'.join(modules)

    assert name
    """
    if not name:
        filename, _, _, _ = stack[0]
        name = strip_py(os.path.basename(filename))
    """

    if _debug_caller_module_name: print(f'caller_module_name: {name}') #DEBUG
    return name

def is_package_type(object):
    '''
        Returns True if object is a python package, else False.

        >>> import denova.python
        >>> is_package_type(denova.python)
        True
        >>> import denova.python.utils
        >>> is_package_type(denova.python.utils)
        False
    '''

    # this seems to be roughly what python does internally
    return (is_module_type(object) and
        (os.path.basename(object.__file__).endswith('__init__.py') or
         os.path.basename(object.__file__).endswith('__init__.pyc') or
         os.path.basename(object.__file__).endswith('__init__.pyo')))

def is_module_type(object):
    ''' Returns True if object is a python module, else False.

        Convenience function for symmetry with is_package_type().

        >>> import denova.python
        >>> is_module_type(denova.python)
        True
        >>> import denova.python.utils
        >>> is_module_type(denova.python.utils)
        True
    '''

    return isinstance(object, types.ModuleType)

def is_instance(obj, cls):
    '''
        More reliable version of python builtin isinstance()

        >>> is_instance('denova.python', str)
        True
    '''

    log(f'is_instance(obj={obj}, cls={cls})')
    log(f'is_instance() type: obj={type(obj)}, cls={type(cls)}')
    try:
        mro = obj.__mro__
    except AttributeError:
        mro = type(obj).__mro__
    log(f'is_instance() mro: {mro}')
    match = cls in mro

    log(f'is_instance() match: {match}')
    return match

def is_class_instance(obj):
    ''' Returns whether the object is an instance of any class.

        You can't reliably detect a class instance with

            isinstance(obj, types.InstanceType)

        as of Python 2.6 2013-05-02. The types module only handles old style
        python defined classes, so types.InstanceType only detects instances
        of the same style.

        >>> import datetime
        >>> c_style_class_instance = datetime.date(2000, 12, 1)
        >>> is_class_instance(c_style_class_instance)
        True

        >>> class OldStyleClass:
        ...     class_data = 27
        ...
        ...     def __init__(self):
        ...         self.instance_data = 'idata'

        ...     def instance_function(self):
        ...         return 3
        >>> old_c = OldStyleClass()
        >>> is_class_instance(old_c)
        True

        >>> class NewStyleClass(object):
        ...     class_data = 27
        ...
        ...     def __init__(self):
        ...         self.instance_data = 'idata'

        ...     def instance_function(self):
        ...         return 3
        >>> new_c = NewStyleClass()
        >>> is_class_instance(new_c)
        True

        >>> # base types are not instances
        >>> is_class_instance(2)
        False
        >>> is_class_instance([])
        False
        >>> is_class_instance({})
        False

        >>> # classes are not instances
        >>> is_class_instance(datetime.date)
        False
        >>> is_class_instance(OldStyleClass)
        False
        >>> is_class_instance(NewStyleClass)
        False

        >>> # test assumptions and python imlementation details

        >>> t = type(2)
        >>> str(t) == "<class 'int'>"
        True
        >>> t = type([])
        >>> str(t) == "<class 'list'>"
        True
        >>> t = type({})
        >>> str(t) == "<class 'dict'>"
        True

        >>> cls = getattr(2, '__class__')
        >>> str(cls) == "<class 'int'>"
        True
        >>> superclass = getattr(cls, '__class__')
        >>> str(superclass) == "<class 'type'>"
        True

        >>> t = str(type(datetime.date))
        >>> t == "<class 'type'>"
        True
        >>> t = str(type(c_style_class_instance))
        >>> t == "<class 'datetime.date'>"
        True
        >>> t = repr(datetime.date)
        >>> t == "<class 'datetime.date'>"
        True
        >>> repr(c_style_class_instance)
        'datetime.date(2000, 12, 1)'
        >>> isinstance(c_style_class_instance, types.MethodType)
        False
        >>> hasattr(c_style_class_instance, '__class__')
        True
        >>> '__dict__' in dir(c_style_class_instance)
        False
        >>> cls = c_style_class_instance.__class__
        >>> hasattr(cls, '__class__')
        True
        >>> '__dict__' in dir(cls)
        False
        >>> hasattr(cls, '__slots__')
        False
        >>> cls = getattr(c_style_class_instance, '__class__')
        >>> str(cls) == "<class 'datetime.date'>"
        True
        >>> superclass = getattr(cls, '__class__')
        >>> str(superclass) == "<class 'type'>"
        True

        >>> ok = '__dict__' in dir(old_c)
        >>> ok == True
        True
        >>> hasattr(old_c, '__slots__')
        False

        >>> '__dict__' in dir(new_c)
        True
        >>> hasattr(new_c, '__slots__')
        False

        '''

    type_str = str(type(obj))

    # old style python defined classes
    if type_str == "<class 'instance'>":
        is_instance = True

    # C defined classes
    elif type_str.startswith('<class '):
        # base types don't have a dot
        is_instance =  '.' in type_str

    # new style python defined classes
    elif type_str.startswith('<'):
        # if it has an address, it's an instance, not a class
        is_instance =  ' 0x' in repr(obj)

    else:
        is_instance = False

    return is_instance

    """ does not detect c-style classes e.g. datetime.xyz
    def is_old_style_instance(obj):
        return isinstance(obj, types.InstanceType)

    def is_new_style_instance(obj):
        # http://stackoverflow.com/questions/14612865/how-to-check-if-object-is-instance-of-new-style-user-defined-class
        is_instance = False
        if hasattr(obj, '__class__'):
            cls = obj.__class__
            if hasattr(cls, '__class__'):
                is_instance = ('__dict__' in dir(cls)) or hasattr(cls, '__slots__')
        return is_instance

    return is_new_style_instance(obj) or is_old_style_instance(obj)
    """

def run(sourcecode):
    '''
        Run source code text.

        >>> run('print("hi")')
        hi
    '''

    # magic. bad. but wasted too many hours trying pythonic solutions
    # in python 2.7 importlib doesn't know spec.Specs is the same as dbuild.spec.Specs

    import tempfile

    __, exec_path = tempfile.mkstemp(
        suffix='.py',
        dir=gettempdir())

    log(f'sourcecode:\n{sourcecode.strip()}')

    with open(exec_path, 'w') as exec_file:
        exec_file.write(sourcecode)

    try:
        exec(compile(open(exec_path).read(), exec_path, 'exec'), globals())
    finally:
        os.remove(exec_path)

def import_module(name):
    '''
        Import with debugging

        >>> module_str = str(import_module("denova.os.user"))
        >>> module_str.startswith("<module 'denova.os.user' from ")
        True
        >>> module_str.endswith("denova/os/user.py'>")
        True
    '''

    try:
        log(f'import_module({name})') #DEBUG
        module = importlib.import_module(name)
        log(f'import_module() result: {module}') #DEBUG
    except ImportError as imp_error:
        log(f'unable to import {name}')
        log('ImportError: ' + str(imp_error))
        msg = f'could not import {name}'
        log(msg)
        # find out why
        from denova.os import command
        log(command.run(['python3', '-c', f'import {name}']).stderr)
        raise ImportError(msg)
    return module

def import_file(name, path):
    '''
        Import source file with debugging

        <<< module = import_file(__file__, '/usr/local/lib/python3.7/dist-packages/denova/os/user.py')
        <<< print(str(module))
        <module 'python.py' from '/usr/local/lib/python3.7/dist-packages/denova/os/user.py'>
    '''

    import importlib

    try:
        log(f'import_file({path})') #DEBUG
        # deprecated in python 3.3
        # the 'right' way to do this varies greatly with the specific python version
        # see http://stackoverflow.com/questions/19009932/import-arbitrary-python-source-file-python-3-3
        #     http://bugs.python.org/issue21436
        # the following is undocumented in python 3, and may not work in all versions
        module = importlib.find_loader(name, path)
        log(f'import_file() result: {module}') #DEBUG
    except ImportError as imp_error:
        log(f'unable to import {path}')
        log('ImportError: ' + str(imp_error))
        msg = f'could not import {path}'
        log(msg)
        raise ImportError(msg)

    return module

def stacktrace():
    '''
        Returns a printable stacktrace.

        Contrary to the python docs, python often limits the number of
        frames in a stacktrace. This is a full stacktrace.

        >>> text = stacktrace()
        >>> assert text.strip().startswith('Traceback ')
        >>> assert 'utils.py' in text
    '''

    s = io.StringIO()
    s.write('Traceback (most recent call last):\n')
    traceback.print_stack(file=s)
    return s.getvalue()

def last_exception(noisy=False):
    ''' Returns a printable string of the last exception.

        If noisy=True calls say() with last_exception_only(). '''

    from denova.python.utils import say

    if noisy:
        say(last_exception_only())
    return traceback.format_exc()

def last_exception_only():
    ''' Returns a printable string of the last exception without a traceback. '''

    type, value, traceback = sys.exc_info()
    if type:
        s = str(type).split('.')[-1].strip('>').strip("'")
        if value is not None and len(str(value)):
            s += f': {value}'
    else:
        s = ''
    return s

def get_module(name):
    ''' Get the module based on the module name.

        The module name is available within a module as __name__.

        >>> module_str = str(get_module(__name__)) # doctest: +ELLIPSIS
        >>> module_str.startswith("<module 'denova.python.utils' from ") or module_str.startswith("<module 'utils' from ")
        True
        >>> module_str.endswith("utils.py'>")
        True
    '''

    return sys.modules[name]

def caller_dir():
    ''' Get the caller's dir.

        This is actually the source dir for the caller of the caller of this module.
    '''

    stack = traceback.extract_stack()[:-2]
    (filename, line_number, function_name, text) = stack[0]
    return os.path.dirname(filename) or os.getcwd()

def caller_file():
    ''' Get the caller's file.

        This is actually the source file for the caller of the caller of this module.
    '''

    stack = traceback.extract_stack()[:-2]
    (filename, line_number, function_name, text) = stack[0]
    return filename

def format_exception(exc):
    ''' Format exception for printing.

        Bug: If exc is not the most recent, but is the same value
        as the most recent exception, we print the most recent traceback.
        This probably will almost never happen in practice.
    '''

    exc_type, exc_value, exc_traceback = sys.exc_info()
    # bug: if exc is not the most recent, but is the same exc_value
    if exc_value == exc:
        exc_str = traceback.format_exc()

    else:
        # the exc Exception is not the most recent, so no traceback
        log('log.debug() called with an exception but no traceback available')
        exc_str = traceback.format_exception_only(type(exc), exc)

    return exc_str

def is_string(obj):
    '''
        Return True iff obj is a string.

        >>> is_string('test')
        True
    '''

    return isinstance(obj, str)

def is_list(obj):
    '''
        Return True iff obj is a list.

        >>> is_list([])
        True
    '''

    log(type(obj))
    return isinstance(obj, list)


def is_tuple(obj):
    '''
        Return True iff obj is a tuple.

        >>> is_string('test')
        True
    '''

    return isinstance(obj, tuple)

def is_dict(obj):
    '''
        Return True iff obj is a dictionary.

        >>> is_string('test')
        True
    '''

    return isinstance(obj, dict)

def exec_trace(code, ignoredirs=[sys.prefix, sys.exec_prefix], globals=None, locals=None, coverdir='/tmp'):
    ''' Trace code.

        Code must be a string. Code must start without leading spaces in the string.

        exec_trace() usually requires passing "globals=globals(), locals=locals()".

        Example:
            from denova.python.utils import exec_trace
            exec_trace("""
from reinhardt.events import log_event
log_event(name, request=request, details=details)
                """,
                globals=globals(), locals=locals())
        '''

    tracer = trace.Trace(ignoredirs=ignoredirs)
    tracer.runctx(code.strip(), globals=globals, locals=locals)
    r = tracer.results()
    r.write_results(show_missing=True, coverdir=coverdir)

def pdb_break():
    ''' Breakpoint for pdb command line debugger.

    Usage:
        from denova.python.utils import pdb_break ; pdb_break()
    '''

    import pdb
    log('breakpointing for pdb')
    pdb.set_trace()

def winpdb_break():
    ''' Breakpoint for winpdb debugger.

        Example:
            from denova.python.utils import winpdb_break; winpdb_break() #DEBUG
    '''

    import rpdb2 #DEBUG
    log('breakpointing for winpdb')
    rpdb2.start_embedded_debugger("password") #DEBUG

def to_bytes(source, encoding=None):
    ''' Convenience method to convert to bytes. '''

    # if source is a string, encoding is required
    # in other cases it doesn't hurt
    if encoding is None:
        encoding = locale.getpreferredencoding(False)

    return bytes(source, encoding=encoding)

def get_scheme_netloc(url):
    ''' Return (scheme, netloc) from url.

        If the port is non-standard, the netloc is 'domain:port'. Otherwise
        netloc is the domain.

       This is used because python 2.4 and 2.5
       give slightly different results from urlparse.

        >>> get_scheme_netloc('http://DeNova.com')
        ('http', 'DeNova.com')
        >>> get_scheme_netloc('https://test:8211')
        ('https', 'test:8211')
    '''

    parsed_url = urlparse(url)

    try:
        scheme = parsed_url.scheme
        netloc = parsed_url.netloc
    except:   # 'bare except' because it catches more than "except Exception"
        scheme = parsed_url[0]
        netloc = parsed_url[1]

    return (scheme, netloc)

def get_remote_ip(request):
    '''Get the remote ip. If there is a forwarder, assume the first IP
       address (if there are more than 1) is the original machine's address.

       Otherwise, use the remote addr.

       Any errors, return 0.0.0.0
    '''

    Unknown_IP = '0.0.0.0'

    if request:
        try:
            # if we're using a reverse proxy, the ip is the proxy's ip address
            remote_addr = request.META.get('REMOTE_ADDR', '')
            forwarder = request.META.get('HTTP_X_FORWARDED_FOR', '')
            if forwarder and forwarder is not None and len(forwarder) > 0:
                m = re.match('(.*?),.*?', forwarder)
                if m:
                    remote_ip = m.group(1)
                else:
                    remote_ip = forwarder
            else:
                remote_ip = remote_addr

            if not remote_ip or remote_ip is None or len(remote_ip) <= 0:
                remote_ip = Unknown_IP
        except:   # 'bare except' because it catches more than "except Exception"
            log(traceback.format_exc())
            remote_ip = Unknown_IP
    else:
        remote_ip = Unknown_IP
        log('no request so returning unknown ip address')

    return remote_ip

def say(message):
    ''' Speak a message.

        Runs a "say" program, passing the message on the command line.
        Because most systems are not set up for speech, it is not an
        error if the "say" program is missing or fails.

        It is often easy to add a "say" program to a system. For example,
        a linux system using festival for speech can use a one line script:
            festival --batch "(SayText \"$*\")"

        Depending on the underlying 'say' command's implementation, say()
        probably does not work unless user is in the 'audio' group.

        >>> say('test say')
        '''

    enabled = True

    if enabled:
        try:
            from denova.os import command
            # the words are unintelligible, and usually all we want is to know something happened
            # message = 'tick' # just a sound #DEBUG
            # os.system passes successive lines to sh
            message = message.split('\n')[0]
            command.run('say', *message)
        except:   # 'bare except' because it catches more than "except Exception"
            pass

def generate_password(max_length=25, punctuation_chars='-_ .,!+?$#'):
    '''
        Generate a password.

        >>> len(generate_password())
        25
    '''

    # the password must be random, but the characters must be valid for django
    password = ''
    while len(password) < max_length:
        new_char = os.urandom(1)
        try:
            new_char = new_char.decode()
            # the character must be a printable character
            if ((new_char >= 'A' and new_char <= 'Z') or
                (new_char >= 'a' and new_char <= 'z') or
                (new_char >= '0' and new_char <= '9') or
                (new_char in punctuation_chars)):

                # and the password must not start or end with a punctuation
                if (new_char in punctuation_chars and
                    (len(password) == 0 or (len(password) + 1) == max_length)):
                    pass
                else:
                    password += new_char
        except:   # 'bare except' because it catches more than "except Exception"
            pass

    return password

def clean_pathname(pathname):
    ''' Clean a pathname by removing all invalid chars.

        See http://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-pathname-in-python
        From that page, roughly:

            The unicodedata.normalize call replaces accented characters
            with the unaccented equivalent, which is better than simply
            stripping them out. After that all disallowed characters are
            removed. Doesn't avoid possible disallowed pathnames."
    '''

    ascii_pathname = unicodedata.normalize('NFKD', pathname.decode()).encode('ASCII', 'ignore')
    return ''.join(c for c in ascii_pathname if c in valid_pathname_chars)


def strip_spaces_and_blank_lines(data):
    '''
        Strip leading and trailing spaces plus blank lines.

        >>> data = "  This is the first line.  \\n   \\n\\n This is the second line.\\n"
        >>> strip_spaces_and_blank_lines(data)
        'This is the first line.\\nThis is the second line.\\n'
    '''

    try:
        if data is not None:
            new_data = []
            for line in data.split('\n'):
                line = trim(line, ' ').decode()
                if len(line.strip('\n')) > 0:
                    new_data.append(f'{line}\n')

            data = ''.join(new_data)
    except:   # 'bare except' because it catches more than "except Exception"
        print(traceback.format_exc())

    return data

def pipe(value, *fns):
    ''' Pipe data from functions a() to b() to c() to d() ...

        "pipe(x, a, b, c, d)" is more readble than "d(c(b(a(x))))".

        See http://news.ycombinator.com/item?id=3349429

        pipe() assumes every function in its list will consume and return the data.
        If you need more control such as filtering and routing, see
        the DeNova's coroutine package.

        >>> def sqr(x):
        ...     return x*x

        >>> def half(x):
        ...     return x/2.0

        >>> for i in range(5):
        ...     pipe(i, sqr, half)
        0.0
        0.5
        2.0
        4.5
        8.0
    '''

    for fn in fns:
        value = fn(value)
    return value

def ltrim(string, prefix):
    ''' Trim all prefixes from string. '''

    length = len(prefix)
    while string.startswith(prefix):
        string = string[length:]
    return string

def rtrim(string, suffix):
    ''' Trim all suffixes from string. '''

    length = len(suffix)
    while string.endswith(suffix):
        string = string[:-length]
    return string

def trim(string, xfix):
    ''' Trim all prefixes or suffixes of xfix from string. '''

    if is_string(string):
        string = string.encode()
    if is_string(xfix):
        xfix = xfix.encode()

    length = len(xfix)
    while string.startswith(xfix):
        string = string[length:]
    while string.endswith(xfix):
        string = string[:-length]
    return string

def remove_lines(string, count):
    ''' Remove lines from string.

        If count is negative, removes lines from end of string. '''

    if count > 0:
        string = '\n'.join(string.split('\n')[count:])
    elif count < 0:
        string = '\n'.join(string.split('\n')[:count])
    return string

def pathmatch(path, pattern):
    ''' Test whether the path matches the pattern.

        This is a mess that needs to be replaced with an ant-style path match.

        The pattern is a shell-style wildcard, not a regular expression.
        fnmatch.fnmatch tests filenames, not paths.

        '**' at the beginning of a pattern matches anything at the beginning
        of a path, but no other wildcards are allowed in the pattern. '''

    def split(path):
        path = os.path.expanduser(path)
        path = os.path.abspath(path)
        return path.split('/')

    if pattern.startswith('**'):
        result = path.endswith(pattern[2:])

    else:
        path = split(path)
        pattern = split(pattern)
        result = (len(path) == len(pattern) and
            all(fnmatch(path[i], pattern[i]) for i in range(len(path))))

    return result

def resolve_path(path):
    ''' Resolves file path wildcards, links, and relative directories.

        To resolve a wildcard path that matches more than one file, use
        glob() and pass each result to resolve_path().

        Returns None if wildcard does not match any files. Raises
        ValueError if wildcard matches more than one file. '''

    paths = glob(path)
    if paths:
        if len(paths) > 1:
            raise ValueError(f'Matches more than one path: {path}')
            path = os.path.normpath(os.path.realpath(paths[0]))
    else:
        path = None
    return path

def domain_base(domain):
    ''' Returns base name from domain.

        I.e. base.tld or base.co.countrydomain or base.com.countrydomain
        all have the same base name.

        E.g. google.com, google.bg, google.de, google.co.in all are based
        on google.

        This can be fooled by domain spoofers or squatters.

        >>> domain_base('denova.com')
        'denova'
    '''

    if type(domain) is bytes:
        domain = domain.decode()
    # regexes might be clearer (or not) but would be slower
    parts = domain.split('.')
    if len(parts) > 1:
        # toss the top level domain
        parts = parts[:-1]
        if len(parts) > 1:
            # toss generic second level domains
            if parts[-1] in ['com', 'co', 'org', 'net']:
                parts = parts[:-1]
    # top level left is the base name
    return parts[-1]

class textfile(object):
    ''' Open possibly gzipped text file as file using contextmanager.

        E.g. "with textfile('mytext.gz') as f".

        Avoids "AttributeError: GzipFile instance has no attribute '__exit__'"
        prior to Python 3.1.

        As of Python 2.6 contextlib.closing() doesn't work. It doesn't expose underlying
        gzip functions because its __enter__() returns the inner object, and it has no
        __getattr__()
        to expose the inner gzip.open(). '''

    def __init__(self, filename, rwmode='r'):
        if filename.endswith('.gz'):
            self.f = gz.open(filename, f'{rwmode}b')
        elif filename.endswith('.bz2'):
            self.f = bz2.BZ2File(filename, f'{rwmode}b')
        elif filename.endswith('.zip'):
            self.f = zipfile.ZipFile(filename, f'{rwmode}b')
        else:
            self.f = open(filename, rwmode)
        self.opened = True

    def __iter__(self):
        return iter(self.f)

    def __enter__(self):
        return self.f

    def __exit__(self, *exc_info):
        self.close()

    def unused_close(self):
        if self.opened:
            self.f.close()
            self.opened = False

    def __getattr__(self, name):
        return getattr(self.f, name)

def gzip(uncompressed):
    ''' Gzip a string '''

    compressed_fileobj = StringIO()
    with gz.GzipFile(fileobj=compressed_fileobj, mode='w') as f:  #, compresslevel=5) as f:
        f.write(uncompressed)
    return compressed_fileobj.getvalue()

def gunzip(compressed):
    ''' Gunzip a string '''

    compressed_fileobj = StringIO(compressed)
    with gz.GzipFile(fileobj=compressed_fileobj, mode='r') as f:
        uncompressed = f.read()
    return uncompressed

@contextmanager
def chdir(dirname=None):
    ''' Chdir contextmanager that restores current dir.

        From http://www.astropython.org/snippet/2009/10/chdir-context-manager

        This context manager restores the value of the current working
        directory (cwd) after the enclosed code block completes or
        raises an exception.  If a directory name is supplied to the
        context manager then the cwd is changed prior to running the
        code block.
    '''

    curdir = os.getcwd()
    try:
        if dirname is not None:
            os.chdir(dirname)
        yield
    finally:
        os.chdir(curdir)

def different(file1, file2):
    ''' Returns whether the files are different. '''

    # diff succeeds if there is a difference, and fails if no difference
    try:
        from denova.os import command
        command.run('diff', file1, file2, brief=True)
        different = False

    except CalledProcessError:
        different = True

    return different

def slugify(value):
    ''' Converts string to a form usable in a url withjout encoding.

        Strips white space from ends, converts to lowercase,
        converts spaces to hyphens, and removes non-alphanumeric characters.
    '''
    value = value.strip().lower()
    value = re.sub('[\s-]+', '-', value)

    newvalue = ''
    for c in value:
        if (
              (c >= 'A' and c <= 'Z') or
              (c >= 'a' and c <= 'z') or
              (c >= '0' and c <= '9') or
              c == '-' or
              c == '_'
              ):
            newvalue += c
    return newvalue

def replace_strings(text, replacements, regexp=False):
    """ Replace text. Returns new text.

        'replacements' is a dict of {old: new, ...}.
        Every occurence of each old string is replaced with the
        matching new string.

        If regexp=True, the old string is a regular expression.

        >>> text = 'ABC DEF 123 456'
        >>> replacements = {
        ...     'ABC': 'abc',
        ...     '456': 'four five six'
        ... }
        >>> replace_strings(text, replacements)
        'abc DEF 123 four five six'
    """

    for old in replacements:
        new = replacements[old]
        if regexp:
            text = re.sub(old, new, text)
        else:
            text = text.replace(old, new)
    return text

def delete_empty_files(directory):
    ''' Delete empty files in directory.

        Does not delete any subdirectories or files in them.

        >>> from tempfile import mkdtemp, mkstemp
        >>> directory = mkdtemp()
        >>> assert os.path.isdir(directory)
        >>> handle, filename1 = mkstemp(dir=directory)
        >>> os.close(handle)
        >>> assert os.path.exists(filename1)
        >>> handle, filename2 = mkstemp(dir=directory)
        >>> os.close(handle)
        >>> assert os.path.exists(filename2)
        >>> with open(filename2, 'w') as f2:
        ...     len = f2.write('data')
        >>> delete_empty_files(directory)
        >>> assert not os.path.exists(filename1)
        >>> assert os.path.exists(filename2)
        >>> os.remove(filename2)
        >>> assert not os.path.exists(filename2)
        >>> os.rmdir(directory)
        >>> assert not os.path.isdir(directory)
    '''

    wildcard = os.path.join(directory, '*')
    for filename in glob(wildcard):
        if os.path.getsize(filename) <= 0:
            os.remove(filename)

def randint(min=None, max=None):
    ''' Get a random int.

        random.randint() requires that you specify the min and max of
        the integer range for a random int. But you almost always want
        the min and max to be the system limits for an integer.
        If not use random.randint().

        'min' defaults to system minimum integer.
        'max' defaults to system maximum integer.
    '''

    import random

    maxsize = sys.maxsize

    if min is None:
        min = -(maxsize-1)
    if max is None:
        max = maxsize

    return random.randint(min, max)

def strip_youtube_hash(filename):
    ''' Strip youtube hash from end of filename '''

    if '.' in filename:
        rootname, _, extension = filename.rpartition('.')
        youtube_match = re.match(r'(.*)-[a-zA-Z0-9\-_]{11}$', rootname)
        if youtube_match:
            cleanroot = youtube_match.group(1)
            filename = cleanroot + '.' + extension
    return filename

def simplify_episode_filename(filename):
    ''' Strip junk from end of tv episode filename '''

    if '.' in filename:
        rootname, _, extension = filename.rpartition('.')
        cleanroot = re.sub(r'(.*)[Ss](\d\d)[Ee](\d\d).*', r'\1S\2E\3', rootname)
        filename = cleanroot + '.' + extension
    return filename

@contextmanager
def os_environ_context(environ):
    """ Context manager to restore os environment variables. """

    old_environ = dict(os.environ)
    os.environ.update(environ)

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)


if __name__ == "__main__":
    import doctest
    doctest.testmod()




#!/usr/bin/env python3
'''
    Run a command using python subprocess.

    Copyright 2018-2020 DeNova
    Last modified: 2020-10-20
'''

import os
import shlex
import subprocess
import sys
from glob import glob

log = None

def run(*command_args, **kwargs):
    ''' Run a command line.

        Much simpler than using python subprocess directly.

        Returns subprocess.CompletedProcess, or raises
        subprocess.CalledProcessError.

        Example::

            >>> result = run('echo word1 word2')
            >>> result.stdout.strip()
            'word1 word2'

        Or if you prefer less chance of parsing errors::
            >>> result = run('echo', 'word1', 'word2')
            >>> result.stdout.strip()
            'word1 word2'

        By default run() captures stdout and stderr. It returns a
        subprocess.CompletedProcess with stdout and stderr, plus a
        combined stderrout as a string.

        You can also send stdout and stderr to the system stdout/stderr,
        to files, and to file objects. Use run_verbose() if you want standard
        stdout/stderr.

        To send stderr to the system's usual output::

            # this isn't a doctest because in a doctest sys.stdout is a
            # doctest._SpoofOut object.

            import sys
            ...
            run('echo', 'stdout', 'text', stdout=sys.stdout)

        To send stderr to a file::

            >>> with open('/tmp/syr.command.stderr', 'w') as stderr:
            ...     result = run('ls', '/tmp', stderr=stderr)

        If you redir stdout or stderr to a file, the .stdout and .stderr
        values in the result returned from run() are None.

        Each command line arg should be a separate run() arg so
        subprocess.check_output can escape args better.

        Unless output_bytes=True, the .stdout, and .stderr
        attributes of CompletedProcess are returned as unicode strings, not
        bytes. For example stdout is returned as stdout.decode().strip(). The
        default is output_bytes=False. This is separate from
        universal_newlines processing, and does not affect stdin.

        Unless glob=False, args are globbed.

        Except for 'output_bytes' and 'glob', keyword args are passed to subprocess.run().

        On error raises subprocess.CalledProcessError.
        The error has an extra data member called 'output' which is a
        string containing stderr and stdout.
    '''

    """ Because we are using PIPEs, we need to use Popen() instead of
        run(), and call Popen.communicate() to avoid zombie processes.
        But to get run()'s timeout, input and check params, we don't.
        Zombie processes are worrisome, but do no real harm.

        See https://stackoverflow.com/questions/2760652/how-to-kill-or-avoid-zombie-processes-with-subprocess-module
    """

    def format_output(result):
        if (result.stdout is not None) and (not isinstance(result.stdout, str)):
            result.stdout = result.stdout.decode()
            result.stdout = result.stdout.strip()
        if (result.stderr is not None) and (not isinstance(result.stderr, str)):
            result.stderr = result.stderr.decode()
            result.stderr = result.stderr.strip()
        return result

    _init_log()
    result = None

    # if there is a single string arg with a space, it's a command line string
    if len(command_args) == 1 and isinstance(command_args[0], str) and ' ' in command_args[0]:
        # run() is better able to add quotes correctly when each arg is separate
        command_args = shlex.split(command_args[0])

        recommended_args = []
        for arg in command_args:
            recommended_args.append(f"'{arg}'")
        log.warning(f"run({', '.join(recommended_args)}) is recommended to avoid command line parsing errors")

        # apparently subprocess.run(..., shell=True) calls shlex.split()
        # kwargs['shell'] = True

    try:
        command_str = ' '.join(map(str, command_args))

        if 'glob' in kwargs:
            globbing = kwargs['glob']
            del kwargs['glob']
        else:
            globbing = True

        # subprocess.run() wants strings
        args = []
        for arg in command_args:
            arg = str(arg)
            if globbing and ('*' in arg or '?' in arg):
                args.extend(glob(arg))
            else:
                args.append(arg)

        if 'output_bytes' in kwargs:
            output_bytes = kwargs['output_bytes']
            del kwargs['output_bytes']
            log(f'output bytes: {output_bytes}')
        else:
            output_bytes = False

        for output in ['stdout', 'stderr']:
            if output not in kwargs:
                kwargs[output] = subprocess.PIPE

        """
        # how to print program's stderr
        # this also handles unicode encoded bytestreams better

        class CompletedProcessStub:
            pass

        proc = subprocess.Popen(proc_args,
                                **kwargs)

        # stderr to the console's stdout
        proc_stderr = ''
        err_data = proc.stderr.readline()
        while err_data:
            line = err_data.decode()
            proc_stderr = proc_stderr + line
            # lines already have a newline
            print(line, end='')
            err_data = proc.stderr.readline()

        # get any stdout from the proc
        proc_stdout, _ = proc.communicate()

        result = CompletedProcessStub()
        result.resultcode = proc.wait()
        result.stdout = proc_stdout
        result.stderr = proc_stderr

        # much simpler code if we don't need realtime feedback
        """
        result = subprocess.run(args,
                                check=True,
                                **kwargs)

    except subprocess.CalledProcessError as cpe:
        log(f'failed command. returncode: {cpe.returncode} / {command_str}')

        # add .stderrout data member
        cpe.stderrout = None
        if cpe.stderr and cpe.stdout:
            log('have cpe.stderr and cpe.stdout') # DEBUG
            cpe.stderrout = (cpe.stderr.decode().strip() +
                          '/n' +
                          cpe.stdout.decode().strip())
        elif cpe.stderr:
            log('have just cpe.stderr') # DEBUG
            cpe.stderrout = cpe.stderr.decode().strip()
        elif cpe.stdout:
            log('have just cpe.stdout') # DEBUG
            cpe.stderrout = cpe.stdout.decode().strip()
        if cpe.stderrout:
            log(cpe.stderrout)

        raise

    except Exception as e:
        log(f'command got Exception: {command_args}')
        log(f'error NOT subprocess.CalledProcessError: {type(e)}')
        log(e)
        raise

    else:
        result = format_output(result)

    return result

def run_verbose(*args, **kwargs):
    ''' Run program with stdout and stderr directed to
        sys.stdout and sys.stderr.
    '''

    result = run(*args, stdout=sys.stdout, stderr=sys.stderr, **kwargs)

    return result

def background(*command_args, **kwargs):
    ''' Run a command line in background.

        If your command file is not executable and starts with '#!',
        background() will use 'shell=True'.

        Passes along all keywords to subprocess.Popen().
        That means you can force the subprocess to run in the foreground
        with e.g. timeout= or check=. But command.run() is a better
        choice for this case.

        Returns a subprocess.Popen object, or raises
        subprocess.CalledProcessError.

        If you redirect stdout/stderr, be sure to catch exceution errors:

            >>> try:
            ...     program = background('python', '-c', 'not python code',
            ...                          stdout=sys.stdout,
            ...                          stderr=sys.stderr)
            ...     wait(program)
            ... except Exception as exc:
            ...     print(exc)
            argument of type 'NoneType' is not iterable

        The caller can simply ignore the return value, poll() for when
        the command finishes, wait() for the command, or communicate()
        with it.

        Do not use Popen.wait(). Use denova.os.command.wait().
    '''

    if not command_args:
        raise ValueError('missing command')

    _init_log()

    # if there is a single string arg with a space, it's a command line string
    if len(command_args) == 1 and isinstance(command_args[0], str) and ' ' in command_args[0]:
        # run() is better able to add quotes correctly when each arg is separate
        command_args = shlex.split(command_args[0])

    command_str = ' '.join(command_args)
    kwargs_str = ''
    for key in kwargs:
        kwargs_str = kwargs_str + f', {key}={kwargs[key]}'

    try:
        process = subprocess.Popen(command_args, **kwargs)

    except OSError as ose:
        log.debug(f'command: {command_str}')
        log.debug(f'OSError as ose: {ose} ({dir(ose)})')
        log.debug(ose)

        if 'Exec format error' in ose.strerror:
            # if the program file starts with '#!' retry with 'shell=True'.
            program_file = command_args[0]
            with open(program_file) as program:
                first_chars = program.read(2)
                if str(first_chars) == '#!':

                    return subprocess.Popen(command_args, shell=True, **kwargs)

                else:
                    log.debug(f'no #! in {program_file}')
                    raise

        else:
            raise

    except Exception as e:
        log.debug(f'command: {command_str}')
        log.debug(e)
        raise

    else:
        log.debug(f"background process started: \"{' '.join(process.args)}\", pid: {process.pid}")
        return process

def wait(program):
    ''' Wait for a background command to finish.'''

    if not isinstance(program, subprocess.Popen):
        raise ValueError('program must be an instance of subprocess.Popen')

    os.waitpid(program.pid, 0)

def _init_log():
    ''' Initialize log. '''

    global log

    if log is None:
        # log import delayed to avoid recursive import.
        from denova.python.log import get_log
        log = get_log()


if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
    Create variables within templates

    Last modified: 2020-10-22

    Variables are set at render time. They are not available earlier,
    such as when {% block %} is evaluated.

    From http://djangoclips.org/clips/829/:

        Here is a Django template tag that allows you to create complex
        variables specified in JSON format within a template.

        It enables you to do stuff like:

        {% var as person %}
        {
             "firstName": "John",
             "lastName": "Smith",
              "address": {
                  "streetAddress": "21 2nd Street",
                  "city": "New York",
                  "state": "NY",
                  "postalCode": 10021
              },
              "phoneNumbers": [
                  "212 555-1234",
                  "646 555-4567"
              ]
          }
         {% endvar %}

         <p>{{person.firstName}}, </br>
            {{person.address.postalCode}}, </br>
            {{person.phoneNumbers.1}}
         </p>

        This tag also enables me to do dynamic CSS using as follows:

        # urlpatters
        urlpatterns = [
            (r'^css/(?P<path>.*\.css)$', 'get_css'),
        ]

        #views
        def get_css(request, path):
            return render(request, 'css/%s' % path, {},
                mimetype="text/css; charset=utf-8")

        # dynamic css within in /path/to/app/templates/css'
        {% load var %}
        {% var as col %}
        {
            "darkbg": "#999",
            "lightbg": "#666"
        }
        {% endvar %}

        {% var as dim %}
        {
            "thinmargin": "2em",
            "thickmargin": "10em"
        }
        {% endvar %}

        body {
            background: {{col.darkbg}};
            margin: {{dim.thinmargin}};
        }
'''

try:
    from django import template
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

import json
import re

from denova.python.log import get_log

log = get_log()

register = template.Library()

class VariablesNode(template.Node):
    def __init__(self, nodelist, var_name):
        self.nodelist = nodelist
        self.var_name = var_name
        log.debug(f'var_name: {var_name}')

    def render(self, context):
        source = self.nodelist.render(context)
        if source.strip().startswith('{'):
            value = json.loads(source)
            #log.debug('value from json: {}'.format(value))
        else:
            value = source
            #log.debug('value from source: {}'.format(value))
        log.debug(f'value type: {type(value)}')
        context[self.var_name] = value
        return ''

@register.tag(name='var')
def do_variables(parser, token):
    try:
        tag_name, arg = token.contents.split(None, 1)
    except ValueError:
        msg = f'"{token.contents.split()[0]}" tag requires arguments'
        log.debug(msg)
        raise template.TemplateSyntaxError(msg)
    log.debug(f'arg: {arg}')
    m = re.search(r'as (\w+)', arg)
    if m:
        var_name, = m.groups()
        log.debug(f'var_name arg: {var_name}')
    else:
        msg = f'"{tag_name}" tag had invalid arguments'
        log.debug(msg)
        raise template.TemplateSyntaxError(msg)

    nodelist = parser.parse(('endvar',))
    parser.delete_first_token()
    return VariablesNode(nodelist, var_name)



'''
    Net utilities.

    Copyright 2014-2020 DeNova
    Last modified: 2020-10-20

    There is some inconsistency in function naming.

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

import asyncio
import json
import re
import socket
from traceback import format_exc
from contextlib import contextmanager
from ssl import SSLContext, PROTOCOL_TLS
from urllib.parse import urlencode
from urllib.request import build_opener, HTTPCookieProcessor, ProxyHandler, Request
from urllib.error import HTTPError, URLError

from denova.os.command import run
from denova.os.user import whoami, require_user
from denova.python.log import get_log
from denova.python.utils import stacktrace

log = get_log()

class NetException(Exception):
    pass

def hostname():
    ''' Convenience method to get the host name.

        >>> from_hostname = hostname()
        >>> from_uname = run('uname', '--nodename').stdout.strip()
        >>> assert from_hostname == from_uname, 'hostname(): {}, uname: {}'.format(from_hostname, from_uname)
    '''

    return socket.gethostname()

def hostaddress(name=None):
    ''' Get the host ip address.

        Returns None if not found.
        Default is to return this host's ip.

        Because this function uses gethostbyname(), be sure you are not
        vulnerable to the GHOST attack.
        https://security-tracker.debian.org/tracker/CVE-2015-0235
    '''

    ip = None

    host = name or hostname()
    #log.debug('host: {}'.format(host))

    try:
        host_by_name = socket.gethostbyname(host)

    except socket.gaierror:
        log.debug(f'no address for hostname: {host}')

    else:
        #log.debug('host by name: {}'.format(host_by_name))

        if name:
            ip = host_by_name

        else:
            # socket.gethostbyname(hostname()) can be wrong depending on what is in /etc/hosts
            # but interface_from_ip() requires we are root, and we want to continue if possible
            if whoami() == 'root':
                interface = interface_from_ip(host_by_name)
                if interface and not interface == 'lo':
                    log.debug(f'setting ip to host by name: {host_by_name}')
                    ip = host_by_name
                else:
                    log.warning(f'socket.gethostbyname(hostname()) returned {host_by_name}, ' +
                        'but no interface has that address. Is /etc/hosts wrong?')

            else:
                # accept socket.gethostbyname() because we can't verify it
                ip = host_by_name

    if not ip:
        # use the first net device with an ip address, excluding 'lo'
        for interface in interfaces():
            if not ip:
                if interface != 'lo':
                    ip = device_address(interface)
                    log.debug(f'set ip to {ip} from first net device {interface}')

    if ip:
        log.debug(f'ip address: {ip}')
    else:
        msg = 'no ip address'
        log.debug(msg)
        raise NetException(msg)

    return ip

def interfaces():
    ''' Get net interfaces.

        >>> if whoami() == 'root':
        ...     'lo' in interfaces()
        ... else:
        ...     print(True)
        True
    '''

    require_user('root')
    output = run(*['/sbin/ifconfig']).stdout

    return re.findall(r'^(\S+?):?\s', output, flags=re.MULTILINE)

def net_interface():
    ''' Get the first net device excluding 'lo'.
        This is normally the ethernet interface.

        Return None if none.

        >>> if whoami() == 'root':
        ...     net_interface() != None
        ... else:
        ...     print(True)
        True
    '''

    first = None
    for interface in interfaces():
        if not first:
            if interface != 'lo':
                first = interface
    return first

def device_address(device):
    ''' Get device ip address

        >>> if whoami() == 'root':
        ...     address = device_address('lo')
        ...     address == '127.0.0.1'
        ... else:
        ...     print(True)
        True
    '''

    require_user('root')

    ip = None
    output = run(*['/sbin/ifconfig', device]).stdout
    for line in output.split('\n'):
        if not ip:
            m = re.match(r'.*inet (addr:)?(\d+\.\d+\.\d+\.\d+)\s', line)
            if m:
                ip = str(m.group(2))

    log.debug(f'{device} ip: {ip}')

    return ip

def mac_address(device):
    '''
        Get device mac address

        >>> if whoami() == 'root':
        ...     mac_address('lo')
    '''

    require_user('root')
    mac = None
    output = run(*['/sbin/ifconfig', device]).stdout
    for line in output.split('\n'):
        if not mac:
            m = re.match(r'.* HWaddr +(..:..:..:..:..:..)', line)
            if m:
                mac = m.group(1)
                log.debug(f'mac: {mac}')

    return mac

def interface_from_ip(ip):
    ''' Find interface using ip address

        >>> if whoami() == 'root':
        ...     interface_from_ip('127.0.0.1')
        ... else:
        ...     print("'lo'")
        'lo'
    '''

    interface_found = None
    for interface in interfaces():
        if not interface_found:
            if ip == device_address(interface):
                interface_found = interface

    return interface_found

@contextmanager
def connect_to(host, port):
    ''' Context manager that returns TCP socket connected to the host and port.

        Closes socket on context exit.

        >>> manager = connect_to('127.0.0.1', '80')
        >>> str(manager).startswith('<contextlib._GeneratorContextManager object')
        True
    '''

    s = None

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((host, port))
        yield s

    finally:
        if s:
            s.close()

def is_listening(host, port):
    '''
        Returns True if a listener is open on host:port. Else returns False.

        Attempts to open a connection to host:port, and if successful
        immediately closes it. A timeout returns False.

        >>> is_listening('127.0.0.1', 80)
        True
    '''


    try:
        # log('is_listening() socket.socket()') # DEBUG
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        # log('is_listening() s.connect()') # DEBUG
        s.connect((host, port))
        # log('is_listening() s.close()') # DEBUG
        s.close()
    except Exception:
        listening = False
    else:
        listening = True

    # log.debug('listening on {}: {}'.format(port, listening))

    return listening

def set_etc_hosts_address(hostname, ip):
    '''
        Set host address in /etc/hosts from device address.

    '''

    def read_file(path):
        with open(path) as file:
            contents = file.read()
        return contents

    def write_etc_hosts(text):
        assert text.strip()
        with open('/etc/hosts', 'w') as hosts_file:
            hosts_file.write(text)

    def edit_text():
        # write /etc/hosts

        hostname_found = False
        newlines = []
        for line in oldlines:

            parts = line.split()

            # if hostname is already in /etc/hosts
            if hostname in parts:
                parts[0] = ip
                hostname_found = True

            line = ' '.join(parts)
            log.debug(f'new line: {line}')
            newlines.append(line)

        # if hostname is not in /etc/hosts
        if not hostname_found:
            # append the ip and hostname
            line = f'{ip} {hostname}'
            newlines.append(line)

        newtext = '\n'.join(newlines).strip() + '\n'
        log.debug(f'new text:\n{newtext}')
        return newtext

    require_user('root')

    oldlines = read_file('/etc/hosts').strip().split('\n')
    log.debug('old /etc/hosts:\n{}'.format('\n'.join(oldlines)))

    newtext = edit_text()
    assert newtext
    write_etc_hosts(newtext)

    # check /etc/hosts
    assert read_file('/etc/hosts') == newtext

def require_socks():
    '''
        Require python-socks/socksipy socks module. If missing, tell where to get it (which wasn't obvious).

    '''

    try:
        import socks
    except ImportError:
        pkg = 'python3-socks'
        msg = f'Requires the debian module from {pkg}, or socksipy from pypi. Called from:\n{stacktrace()}'
        log(msg)
        raise NetException(msg)

def send_api_request(url, params, proxy_dict=None, user_agent=None):
    '''
        Send a post to a url and get the response.

        >>> params = []
        >>> page = send_api_request('https://denova.com', params)
        >>> page is not None
        True

        >>> page = send_api_request('https://denova.com/api/', {'api_version': '1.0', 'actions': 'versions'})
        >>> page is not None
        True

        >>> params = []
        >>> page = send_api_request('https://denova.com', params, proxy_dict={'https': '127.0.0.1:8398'})
        >>> page is not None
        True
    '''

    def strip_left(html, search_string):
        i = html.lower().find(search_string)
        if i >= 0:
            html = html[i+len(search_string):]

        return html

    def strip_right(html, search_string):
        i = html.lower().find(search_string)
        if i >= 0:
            html = html[:i]

        return html

    # get the referer if possible
    referer = None
    __, __, ipaddrlist = socket.gethostbyaddr(socket.getfqdn())
    for ip_address in ipaddrlist:
        if ip_address != '127.0.0.1':
            referer = ip_address
            log(f'send_api_request with referer: {referer}')
            break

    page = post_data(url, params, proxy_dict=proxy_dict, user_agent=user_agent, referer=referer)

    if page is None:
        body_html = ''
        log('page is empty')
    else:
        body_html = page.strip()
        body_html = body_html.lstrip()
        body_html = strip_left(body_html, b'<body>')
        body_html = strip_right(body_html, b'</body>')
        body_html = body_html.strip()
        body_html = body_html.lstrip()

    return body_html

def post_data(full_url, params, proxy_dict=None, user_agent=None, referer=None):
    '''
        Send a post to a url and return the data.

        full_url: the url, including the scheme, that you want to receive
        params: any extra parameters to be included in the post's header
        proxy_dict: (optional) formatted as {TYPE_OF_PROXY, PROXY_ADDRESS}
                    for example: proxy_dict={'https': 'http://127.0.0.1:8398'}
        user_agent: (optional) User agent included in the post's header (e.g., Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/536.5 (KHTML, like Gecko) Chrome/19.0.1084.56 Safari/536.5)
        referer: (optional) Referer included in the post's header

        >>> page = post_data('https://denova.com', '')
        >>> page is not None
        True
    '''

    def split_host_url(full_url):
        # extract the host and the remainder of the url
        m = re.match('https?://(.*/?)(.*)', full_url)
        if m:
            host = m.group(1)
            try:
                url = f'/{m.group(2)}'
            except:  # 'bare except' because it catches more than "except Exception"
                url = '/'
        else:
            host = full_url
            url = '/'

        if params is not None and len(params) > 0:
            url = f'{url}?{urlencode(params)}'

        return host, url

    def split_host_port(proxy_dict):
        # extract the proxy's host and port
        m = re.match('(\d+.\d+.\d+.\d+):(\d+)', proxy_dict['https'])
        if m:
            proxy_host = m.group(1)
            proxy_port = int(m.group(2))

        return proxy_host, proxy_port

    page = None
    try:
        try:
            # import late so the rest of module can run even if there's no CookieJar
            from http.cookiejar import CookieJar
            if proxy_dict is None:
                opener = build_opener(HTTPCookieProcessor(CookieJar()))
            else:
                proxy_handler = ProxyHandler(proxy_dict)
                opener = build_opener(proxy_handler, HTTPCookieProcessor(CookieJar()))
        except NameError or ImportError:
            msg = 'Requires http.cookiejar'
            log(msg)
            raise NetException(msg)

        if params is None or len(params) < 1:
            encoded_params = None
        else:
            encoded_params = urlencode(params).encode()
            log(f'encoded params: {encoded_params}') #DEBUG

        request = Request(full_url, encoded_params)
        if user_agent is not None:
            request.add_header('User-Agent', user_agent)
        if referer is not None:
            request.add_header('Referer', referer)

        handle = opener.open(request)
        page = handle.read()

    except HTTPError as http_error:
        page = None
        log(f'full_url: {full_url}')
        log(f'http error: {str(http_error)}')

    except URLError as url_error:
        page = None
        log(f'{url_error} to {full_url}')

    except:  # 'bare except' because it catches more than "except Exception"
        page = None
        log(f'full_url: {full_url}')
        log(format_exc())

    return page

def get_page(full_url, proxy_dict=None):
    '''
        Get a page from the url.

        full_url: the url, including the scheme, that you want to receive
        proxy_dict: (optional) formatted as {TYPE_OF_PROXY, PROXY_ADDRESS}
                    for example: proxy_dict={'https': 'http://127.0.0.1:8398'}

        >>> from denova.net.html_addons import extract_text
        >>> page = get_page('http://denova.com')
        >>> page is not None
        True
    '''

    page = None
    try:
        if proxy_dict is None:
            opener = build_opener()
        else:
            proxy_handler = ProxyHandler(proxy_dict)
            opener = build_opener(proxy_handler)

        request = Request(full_url)
        handle = opener.open(request)
        page = handle.read()

    except HTTPError as http_error:
        page = None
        log(f'full_url: {full_url}')
        log(f'http error: {str(http_error)}')

    except URLError as url_error:
        page = None
        log(f'{url_error} to {full_url}')

    except:  # 'bare except' because it catches more than "except Exception"
        page = None
        log(f'full_url: {full_url}')
        log(format_exc())

    return page

def websocket_send(url, message, secure=True):
    ''' Send a websocket message.

        'url' is a websocket url string. 'message' is a dict including a 'type' field.
        If 'secure' is False, ssl is not verified. The default for 'secure' is True.
    '''

    ''' Async syntax is different for different python versions.
        This is for python 3.5.

        Apparently every thread needs its own asyncio event loop.
        Python only creates one automatically for the main thread.
        See
            https://stackoverflow.com/questions/25063403/python-running-autobahnpython-asyncio-websocket-server-in-a-separate-subproce
    '''

    async def send_json():

        #log.debug('send_json')
        try:
            if secure:
                async with websockets.connect(url) as websocket:
                    await websocket.send(msg_json)
            else:
                context = SSLContext(protocol=PROTOCOL_TLS)
                async with websockets.connect(url, ssl=context) as websocket:
                    await websocket.send(msg_json)

        except NameError as n:
            log.debug(f'{str(n)} in send_json')
            raise

        except Exception as e:
            log.debug(e)
            raise

        #log.debug('exit send_json')

    # import delayed until we need websockets
    # on debian install python3-websockets
    import websockets

    #from denova.python.utils import randint
    #nonce = randint()
    #log.debug("websocket_send('{}', '{}') nonce {}".format(url, message, nonce))

    if type(url) is not str:
        raise TypeError(f'websocket url must be a string, not {type(url)}')
    if type(message) is not dict:
        raise TypeError(f'websocket message must be a dict, not {type(message)}')
    if 'type' not in message:
        raise ValueError("'type' required in websocket message")

    try:
        msg_json = json.dumps(message)

        need_loop = False
        try:
            # assume we have a current event loop
            asyncio.get_event_loop().run_until_complete(send_json())

        except RuntimeError as re:
            if 'There is no current event loop in thread' in str(re):
                need_loop = True
            else:
                log.debug(e)

        if need_loop:
            #log.debug('set current event loop')
            """ Apparently every thread needs its own asyncio event loop.
                See https://stackoverflow.com/questions/25063403/python-running-autobahnpython-asyncio-websocket-server-in-a-separate-subproce
                It may be better to use::
                    websockets.connect(loop=...)
            """
            asyncio_event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(asyncio_event_loop)

            asyncio.get_event_loop().run_until_complete(send_json())

    except NameError as n:
        log.debug(f'{str(n)} in websocket_send')
        raise

    except Exception as e:
        log.debug(e)
        raise

    finally:
        #log.debug("exit websocket_send('{}', '{}') nonce {}".format(url, message, nonce))
        pass

def _test_websocket_send(url, message, secure=True):
    ''' Send a websocket message.

        'url' is a websocket url string. 'message' is a dict including a 'type' field.
        If 'secure' is False, ssl is not verified. The default for 'secure' is True.
    '''

    ''' Async syntax is different for different python versions.
        This is for python 3.5.
    '''

    async def send_json():

        log.debug('send_json')
        try:
            if secure:
                async with websockets.connect(url) as websocket:
                    await websocket.send(msg_json)
            else:
                context = SSLContext(protocol=PROTOCOL_TLS)
                async with websockets.connect(url, ssl=context) as websocket:
                    await websocket.send(msg_json)

        except Exception as e:
            log.debug(e)

        log.debug('exit send_json')

    log.debug('websocket_send')

    # import delayed until we need websockets
    # on debian install python3-websockets
    import websockets
    from denova.python.dict import dictify

    if type(url) is not str:
        raise TypeError(f'websocket url must be a string, not {type(url)}')
    if type(message) is not dict:
        raise TypeError(f'websocket message must be a dict, not {type(message)}')
    if 'type' not in message:
        raise ValueError("'type' required in websocket message")

    try:
        msg_json = json.dumps(dictify(message))
        websockets_run(send_json)

    except Exception as e:
        log.debug(e)

    finally:
        log.debug('exit websocket_send')

@contextmanager
def websocket_receive(function, host, port):
    ''' Context manager to receive websocket messages.

        NOT TESTED

        Example::
            # not a doctest because it is a server, and so runs forever
            # you can run this in a separate thread
            async def echo(websocket, path):
                async for message in websocket:
                    await websocket.send(message)
            websocket_receive(echo, 'localhost', 8888)
    '''

    ''' See:
            https://pypi.python.org/pypi/websockets
    '''

    # import delayed until we need websockets
    # on debian install python3-websockets
    import websockets

    asyncio.get_event_loop().run_until_complete(
        websockets.serve(function, host, port))
    asyncio.get_event_loop().run_forever()

def websockets_run(function):
    ''' Run function in asyncio event loop.

        'function' is a function without parameters to be run by asyncio. Example::
            asyncio.get_event_loop().run_until_complete(function())
        Do not include '()'.
    '''

    ''' Apparently every thread needs its own asyncio event loop.
        Python only creates one automatically for the main thread.
        See
            https://stackoverflow.com/questions/25063403/python-running-autobahnpython-asyncio-websocket-server-in-a-separate-subproce
    '''

    try:
        need_loop = False

        try:
            asyncio.get_event_loop().run_until_complete(function())
        except RuntimeError as re:
            if 'There is no current event loop in thread' in str(re):
                need_loop = True
            else:
                log.debug(re)

        if need_loop:
            log.debug('no current event loop')
            """ Apparently every thread needs its own asyncio event loop.
                See https://stackoverflow.com/questions/25063403/python-running-autobahnpython-asyncio-websocket-server-in-a-separate-subproce
                It may be better to use::
                    websockets.connect(loop=...)
            """
            asyncio_event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(asyncio_event_loop)

            asyncio.get_event_loop().run_until_complete(function())

    except Exception as e:
        log.debug(e)

    finally:
        log.debug('exit websocket_send')

def statsd_send(name, value, mtype, server=None, port=None):
    ''' Send to statsd.

        'name' is the name of the stat.

        'value' is the value.

        'mtype' is the statsd metric type, such as 'c' for a counter.

        'server' and 'port' are the statsd server ip and port. The default is 127.0.0.1:8125.
        8125 is the standard statsd port.

        >>> statsd_send('foo', 3, 'c')
        >>> statsd_send('foo2', 3, 'c')
        >>> statsd_send('foo3', 452, 'g')
    '''

    STATS_D_STANDARD_PORT = 8125

    if server is None:
        server = '127.0.0.1'
    if port is None:
        port = STATS_D_STANDARD_PORT

    message = f'{name}:{value}|{mtype}'
    udp_send(message, server, port)

def udp_send(message, server, port):
    ''' Send a udp packet.

        'message' is the bytes to send. 'server' and 'port' are the destination.
    '''

    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM) # UDP
    sock.sendto(bytes(message, "utf-8"), (server, port))


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#! /usr/bin/python3
'''
    Python logs for humans.
    Requires the safelog package available on PyPI.

    Copyright 2008-2020 DeNova
    Last modified: 2020-10-29

    Documentation at https://denova.com/open_source/safelog/

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.

    >>> import denova.python.log
    >>> log = denova.python.log.get()

    >>> log('logs are easy')

    >>> # log exception
    >>> try:
    ...     raise Exception('log exception and traceback')
    ... except:  # 'bare except' because it catches more than "except Exception"
    ...     log.exception()

    >>> # log exception only
    >>> try:
    ...     raise Exception('log exception only')
    ... except:  # 'bare except' because it catches more than "except Exception"
    ...     log.exception_only()

    Developer notes

    Python gets confused about circular imports.
    It's best for this module to delay imports of other modules that use
    this module. An alternative is to use the sh or subprocess module here
    instead of calling the other module. Another alternative is delayed
    import of denova.python.log in the other module. Example::

        _log = None
        def log(msg):
            global _log
            if not _log:
                from denova.python.log import get
                _log = get()
            _log(message)
        ...
        log(message)

    You can't easily use this module to debug this module, and some modules
    can't import this one, e.g. django's settings.py. There is a tiny
    alternative log module, denova.python._log, which writes to '/tmp/_log...':
        1. Use denova.python.log._debug(..., force=True)
        2. To debug denova.python.log itself, dynamically set it to send its internal
           debugging to  '/tmp/_log...'. Example:

               import denova.python.log
               denova.python.log._DEBUGGING = True # DEBUG

           Then inside denova.python.log:

               _debug(...)

    WARNING: Setting _DEBUGGING to True will destroy isolation of logs for
    different users. This will cause strange seemingly unrelated
    errors and can cost you days of debugging. Leave _DEBUGGING set to
    False except when you're explicitly debugging this module.

    Of course you can use python's built-in logger directly, if the
    current release of logger is reliable. Example::

            import logging
            # since settings.py imports * from this module, settings.py may also use this log
            LOG_FILENAME = 'mydir/settings.log'
            logging.basicConfig(filename=LOG_FILENAME,level=logging.DEBUG,
                                format='%(asctime)s %(name)s %(levelname)s %(message)s')
            try:
                os.chmod(LOG_FILENAME, 0o666)
            except:  # 'bare except' because it catches more than "except Exception"
                pass
            plog = logging.debug

            ...

            plog(...)
'''
import os
import re
import shutil
import smtplib
import socket
import sys
from tempfile import NamedTemporaryFile, gettempdir
from datetime import datetime, timezone
from glob import glob
from traceback import format_exc, format_exception, format_exception_only, format_stack

from denova.os.command import run
import denova.python._log

# constants shared with denova.python.log and logwriter are
# in denova.python.log so they can be imported easily by tools
SAFELOG_HOST = "localhost"
SAFELOG_PORT = 8534
# try to avoid overhead of pickling/json
FIELD_SEPARATOR = '\x01'

# analogous to /var/log
BASE_LOG_DIR = '/var/local/log'

LOGS = {}
_MASTER_LOGS = {}
_USE_MASTER_LOG = True
_RAISE_LOGGIN_ERRORS = False

# internal debugging for denova.python.log
_DEBUGGING = False
_DEBUGGING_LOG_REMOVE_DISABLE = False

WORLD_READWRITE_ACCESS = 0o666

# set these values if you want mail sent if a serious error occurs
alert_from_address = None
alert_to_address = None

# while we're here, set up python basic logging
# it's ok to do this multiple times
# each call to call logging.basicConfig() overrides earlier settings
basic_logging_file = NamedTemporaryFile(
    mode='a',
    prefix=os.path.join(gettempdir(), 'denova.python.default.'),
    suffix='.log',
    delete=True)
os.chmod(basic_logging_file.name, WORLD_READWRITE_ACCESS)

def _debug(message, force=False, filename=None, mode=None):
    if force or _DEBUGGING:
        denova.python._log.log(f'logging.log: {message}', filename=filename, mode=mode)

class _Log():
    ''' Log file.

        >>> from denova.python.log import get_log

        >>> log = get()
        >>> log('log message')

        >>> log2 = get_log('/tmp/testlog.log')
        >>> log('log message 2')

        Logs all messages at python's DEBUG level.
    '''

    MAX_STACK = 1000

    log_dir = None
    filename = None

    def __init__(self,
                 filename=None, dirname=None, group=None,
                 recreate=False, verbose=False, audible=False):
        ''' 'filename' is an explicit filename.
            'dirname' is the dir to use with the default log file basename.
            'group' is the group that wns the lof file. Defaults to the group
            for the current user.
            'recreate' will delete any existing log before use.
            'verbose' prints log entries to stdout. The 'verbose' keyword can
            be overridden for a log entry.
            If audible=True, the command line 'say' program will be called
            with the message.
        '''

        self.filename = filename
        self.dirname = dirname
        self.group = group
        self.user = denova.python._log.whoami()
        self.recreate = recreate
        self.verbose = verbose
        self.audible = audible

        self.pathname = get_log_path(filename=self.filename, dirname=self.dirname)
        self.debugging = False
        self.stdout_stack = []

    def __call__(self, message, verbose=None, audible=None, exception=None):
        ''' Output message to log.debug, optionally also to stdout and audibly.

            If verbose=True, this log message will also be printed to stdout.
            If verbose=False, this message will not printed to stdout, even if
            the log was initialized with verbose=True. Default is None, which
            uses the verbose value passed to Log() or get_log(, if any).

            If audible=True, the command line 'say' program will be called
            with the message. Otherwise audible is similar to verbose. Since
            voice generation can be nearly unintelligible, the message should
            be very short and used infrequently.

            If exception=True, the last exception will be logged.
        '''

        self.debug(message)
        # previously was "self.write(message)". did some caller specify 'DEBUG', 'INFO' etc?

        if not self.is_master():

            if verbose is None:
                verbose = self.verbose
            if verbose:
                print(f'{denova.python._log.timestamp()} {message}')

            if audible is None:
                audible = self.audible
            if audible:
                try:
                    from denova.python.utils import say
                    say(message)
                except:  # 'bare except' because it catches more than "except Exception"
                    pass

            # this is 'exception' passed to __call__()
            if exception:
                self.exception()

    def write(self, message):
        ''' Write message to log.

            write() writes to a log as you would to a file.
            This lets you redirect sys.stdout to the log and log print output
            generated by third party libraries, even if the library uses
            print statements without '>>' or print functions without 'stream='.
        '''

        def notify_webmaster(message):
            ''' Send the message to the webmaster.

                We can't use reinhardt.dmail.send because it imports this module.
            '''

            if alert_from_address is not None and alert_to_address is not None:
                addresses = f'From: {alert_from_address}\nTo: {alert_to_address}'
                msg = f'{addresses}\nSubject: {subject}\n\n{message}\n'
                server = smtplib.SMTP('localhost')
                server.sendmail(alert_from_address, alert_to_address, msg)

        # we don't use @synchronized here because of import errors

        global _RAISE_LOGGIN_ERRORS

        try:
            from denova.python.utils import is_string

            if is_string(message):
                self._write(message)
            elif isinstance(message, (bytes, bytearray)):
                # send bytes to self._write()
                self._write(message.decode(errors='replace'))
            else:
                self.write(f'unable to write message because it is a {type(message)}')

            if self.verbose:
                print(message)
            if _USE_MASTER_LOG and not self.is_master():
                if self.user not in _MASTER_LOGS:
                    _MASTER_LOGS[self.user] = get_log('master.log', dirname=self.dirname)

                try:
                    logline = f'{os.path.basename(self.pathname)} {message}'
                    _MASTER_LOGS[self.user]._write(logline)
                except UnicodeDecodeError:
                    _MASTER_LOGS[self.user]._write(f'{self.filename} - !! Unable to log message -- UnicodeDecodeError !!')

        except UnicodeDecodeError:
            try:
                _debug(message.decode(errors='replace'))
            except UnicodeDecodeError:
                self.write(f'unable to write message because it is a type: {type(message)}')
                subject = '!! Unable to log message !!'
                self._write(subject)
                if _RAISE_LOGGIN_ERRORS:
                    _RAISE_LOGGIN_ERRORS = False
                    notify_webmaster(message)

        except Exception:
            if _DEBUGGING:
                sys.exit('debug exit after exception') # DEBUG
            """
            self._write(format_exc())
            subject = '!! Unable to log message !!'
            self._write(subject)
            if _RAISE_LOGGIN_ERRORS:
                _RAISE_LOGGIN_ERRORS = False
                notify_webmaster(message)
            """
            raise

        #if _DEBUGGING: sys.exit('debug exit without exception') # DEBUG

    def _write(self, message):
        ''' Send message to log file. '''

        try:
            # _debug('_write() self._write') # DEBUG
            log_entry = f'{self.timestamp()} {message}\n'
            # for some reason nulls are getting into the logs
            log_entry.replace('\0', '')
            # we use FIELD_SEPARATOR as a separator later,
            # so we don't want them in the data
            log_entry.replace(FIELD_SEPARATOR, '<<FS>>')
            self._write_log_entry(log_entry)

        except:  # 'bare except' because it catches more than "except Exception"
            _debug('_write() after self._write: self.exception') # DEBUG
            self.exception()

    def _write_log_entry(self, log_entry):
        ''' Send line to logserver. '''

        def check_log_entry():
            ok = True
            if FIELD_SEPARATOR in self.user:
                self.warning('denova.python.log: FS in self.user')
                ok = False
            if FIELD_SEPARATOR in basename:
                self.warning('denova.python.log: FS in basename')
                ok = False
            if FIELD_SEPARATOR in log_entry:
                ok = False
                self.warning('denova.python.log: FS in log_entry; binary log messsage?')

            return ok

        def log_line():
            return f'{self.user} {basename} {log_entry.rstrip()}'

        self._check_user()
        basename = os.path.basename(self.pathname)

        # Create a socket (SOCK_STREAM means a TCP socket)
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            # Connect to logserver and send data
            try:
                sock.connect((SAFELOG_HOST, SAFELOG_PORT))

            except ConnectionRefusedError:
                try:
                    msg = f'No log server at {SAFELOG_HOST}:{SAFELOG_PORT}'
                    denova.python._log.log(msg)
                    denova.python._log.log(log_line())
                    sys.exit(msg)

                except:  # 'bare except' because it catches more than "except Exception"
                    # probably system is going down and
                    # the log server is already down
                    pass

            else:
                if check_log_entry():
                    _debug(f'send log entry: {log_line()}')
                    data = FIELD_SEPARATOR.join([self.user, basename, log_entry])
                    data = data.encode()
                    sock.sendall(data)

    def info(self, msg, *args, **kwargs):
        ''' Compatibility with standard python logging. '''

        self.log('INFO', msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        ''' Compatibility with standard python logging.

            If the message is an Exception, Log.debug() logs everything we
            know about the Exception. If you call Log.warning() etc. and
            want Exception details, also call log.debug().
        '''

        if isinstance(msg, UnicodeError):

            # don't log the bad data; it screws up some editors
            self.debug(format_exception_only(type(msg), msg))
            self.debug('UnicodeError:')
            self.stacktrace()

        elif isinstance(msg, Exception):

            if msg.args:
                # log any Exception args
                try:
                    msg.args

                except AttributeError:
                    pass

                else:
                    if isinstance(msg.args, tuple):
                        try:
                            self.debug(' '.join(map(str, list(msg.args))))
                        except UnicodeDecodeError:
                            self.debug(b' '.join(map(bytes, list(msg.args))))
                    else:
                        self.debug(msg.args)

            __, exc_value, __ = sys.exc_info()

            # is the exc_info for this Exception?
            if isinstance(exc_value, type(msg)) and str(exc_value) == str(msg):
                self.stacktrace()
            else:
                # the msg Exception is not the most recent, so no traceback for this exception
                self.debug('log.debug() called with an exception but no traceback available')
                msg = format_exception_only(type(msg), msg)
                self.log('DEBUG', msg)

        else:
            self.log('DEBUG', msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        ''' Compatibility with standard python logging. '''

        self.log('WARNING', msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        ''' Compatibility with standard python logging. '''

        self.log('ERROR', msg, *args, **kwargs)

    def exception(self):
        ''' Log the current exception.

            Any calls to last_exception() should call exception() instead.
        '''

        exc_type, exc_value, exc_traceback = sys.exc_info()
        # if we are in an 'except' clause
        if exc_type:
            try:
                lines = format_exception(exc_type, exc_value, exc_traceback)
                msg = ''.join(lines)
                self.debug(msg)
            except:  # 'bare except' because it catches more than "except Exception"
                print(f'exc_type: {exc_type}, exc_value: {exc_value}, exc_traceback: {exc_traceback}')

    def exception_only(self):
        ''' Log the current exception type and value, without traceback. '''

        exc_type, exc_value, __ = sys.exc_info()
        # if we are in an 'except' clause
        if exc_type:
            try:
                msg = ''
                for line in format_exception_only(exc_type, exc_value):
                    msg = msg + line.strip()
                self.debug(msg)
            except: # 'bare except' because it catches more than "except Exception"
                print(f'exc_type: {exc_type}, exc_value: {exc_value}')

    def log(self, level, msg, *args, **kwargs):
        ''' Compatibility with standard python logging.

            Utility function to support log.debug(), etc.
            Called as self.log('DEBUG', ...), self.log('INFO', ...), etc.
        '''

        if kwargs:
            args = args + (kwargs,)
        if args:
            message = msg.format(*args)
        else:
            message = msg

        try:
            try:
                level.encode('utf-8')
            except UnicodeDecodeError:
                level = ''

            try:
                if isinstance(message, Exception):
                    message = str(message)
                elif isinstance(message, (bytes, bytearray)):
                    message = message.decode(errors='replace')
            except TypeError:
                # don't change message. but would it be better to still encode as utf8?
                pass
            except UnicodeDecodeError:
                message = '-- message contains unprintable characters --'

            self.write(f'{level} {message}')

        except Exception:
            self.exception()

    def stacktrace(self):
        ''' Log full current stacktrace.

            Contrary to the python docs, python often limits the number of
            frames in a stacktrace. This is a full stacktrace.

            >>> import os, os.path
            >>> import time

            >>> TESTLOG = 'test.stacktrace.log'

            >>> log = denova.python.log.get(TESTLOG)

            >>> def test_func():
            ...     log('test stacktrace()')
            ...     log.stacktrace()

            >>> path = get_log_path(TESTLOG)
            >>> if os.path.exists(path):
            ...     os.remove(path)

            >>> test_func()
            >>> # allow time for the logserver to write the stacktrace
            >>> time.sleep(1)
            >>> with open(path) as infile:
            ...     text = infile.read()
            ...     assert 'test_func' in text, text

            Alternate for exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                lines = format_exception(exc_type, exc_value, exc_traceback,
                                         limit=self.MAX_STACK,
                                         chain=True)
                msg = ''.join(lines).strip()
        '''

        lines = ['Stacktrace:\n'] + format_stack()[:-1]
        self.debug(''.join(lines))

    def is_master(self):
        ''' Return whether this log is a master log. '''

        if not self.pathname:
            self.pathname = get_log_path(filename=self.filename, dirname=self.dirname)
        return os.path.basename(self.pathname) == 'master.log'

    def _check_user(self):
        ''' Detect user changed.

            If no dir was specified, use default dir for user.
        '''

        if not self.dirname:
            if not self.filename.startswith('/'):

                current_user = denova.python._log.whoami()

                if self.user is None:
                    self.user = current_user

                elif self.user != current_user:
                    msg = f'Log user changed from {self.user} to {current_user}.'
                    _debug(msg, force=True)
                    self.user = current_user
                    self.pathname = get_log_path(filename=self.filename, dirname=self.dirname)

    def timestamp(self):
        '''
            >>> l = _Log()
            >>> isinstance(l.timestamp(), datetime)
            True
        '''
        try:
            dt = datetime.utcnow().replace(tzinfo=timezone.utc)
        except: # 'bare except' because it catches more than "except Exception"
            dt = datetime.utcnow()
        return dt

    def flush(self):
        ''' This function is here to make _Log
            compatible with other logging system. '''

        # do not delete the pass as it's needed when we strip all comments
        pass

def get(filename=None, dirname=None, group=None, recreate=False, verbose=False):
    ''' Return open log. Default is a log for the calling module.

        The default log path is "BASE_LOG_DIR/USER/MODULE.log".
        If filename is specified and starts with a '/', it is the log path.
        If filename is specified and does not start with a '/', it replaces "MODULE.log".
        If dirname is specified, it replaces "BASE_LOG_DIR/USER".

        Log instances are cached by pathname.

        If recreate=True and the log file is not already open, any existing
        log file is removed.

        >>> import os.path
        >>> import denova.python.log

        >>> log = denova.python.log.get('testlog3.log')
        >>> log('log message')
        >>> user_log_dir = os.path.join(BASE_LOG_DIR, denova.python._log.whoami())
        >>> assert os.path.dirname(log.pathname) == user_log_dir

        >>> log = get('testlog1.log', dirname='/tmp/logs')
        >>> log('log message')
        >>> print(log.dirname)
        /tmp/logs

        >>> log = denova.python.log.get('/tmp/testlog2.log')
        >>> log('log message')
        >>> print(log.dirname)
        /tmp
    '''

    logpath = get_log_path(filename=filename, dirname=dirname)

    if logpath in LOGS.keys():
        log = LOGS[logpath]

    else:
        dirname = os.path.dirname(logpath)
        log = _Log(
            filename=filename, dirname=dirname,
            group=group, recreate=recreate, verbose=verbose)
        LOGS[logpath] = log

    return log

def get_log_path(filename=None, dirname=None):
    ''' Returns log path.

        The default log path is "BASE_LOG_DIR/USER/MODULE.log".
        If filename is specified and starts with a '/', it is the log path.
        If filename is specified and does not start with a '/', it replaces "MODULE.log".

        >>> import os.path
        >>> from denova.python.log import get_log

        >>> path = get_log_path('testlog1.log', dirname='/tmp/logs')
        >>> assert os.path.basename(path) == 'testlog1.log'
        >>> assert os.path.dirname(path) == '/tmp/logs'

        >>> path = get_log_path('/tmp/testlog2.log')
        >>> assert os.path.basename(path) == 'testlog2.log'
        >>> assert os.path.dirname(path) == '/tmp'

        >>> path = get_log_path('testlog3.log')
        >>> assert os.path.basename(path) == 'testlog3.log'
        >>> assert os.path.dirname(path) == os.path.join(BASE_LOG_DIR, denova.python._log.whoami())

        >>> path = get_log_path()
        >>> 'denova.python.log.get_log_path' in path
        True
    '''

    if filename is None:
        from denova.python.utils import caller_module_name
        filename = caller_module_name(ignore=[__file__])
        if filename == '<stdin>':
            filename = '__stdin__'
        else:
            if filename.startswith('..'):
                filename = filename[2:]
            match = re.match('<doctest (.*?)\[', filename)
            if match:
                filename = match.group(1).replace('__main__.', '').lower()

    # sometimes we pass the filename of the caller so strip the end
    if filename.endswith('.py') or filename.endswith('.pyc'):
        filename, _, _ = filename.rpartition('.')

    filename = default_log_filename(filename)
    if dirname is None:
        dirname = default_log_dir()
    else:
        # no relative dir names
        assert dirname.startswith('/')

    logpath = os.path.join(dirname, filename)
    _debug(f'in denova.python.log.get_log_path() logpath: {logpath}') # DEBUG

    return logpath

def default_log_dir():
    '''
       We want to keep the logs together as much as possible,

       The default is /var/local/log/USER. This avoids log ownership conflicts.
    '''

    user = denova.python._log.whoami()
    dirname = os.path.join(BASE_LOG_DIR, user)
    return dirname

def delete_all_logs(dirname=None):
    ''' Delete all files in dir. '''

    if not dirname:
        dirname = default_log_dir()

    if _DEBUGGING_LOG_REMOVE_DISABLE:
        _debug(f'_DEBUGGING_LOG_REMOVE_DISABLE disabled: delete_all_logs() removing {dirname}') #DEBUG

    else:
        # don't delete dirname itself
        entries = glob(os.path.join(dirname, '*'))
        # make sure this dir has at least one log, i.e. is possibly a logs dir
        assert any(entry.endswith('.log') for entry in entries)
        for entry in entries:
            if os.path.isdir(entry):
                shutil.rmtree(entry)
            else:
                os.remove(entry)

def default_log_filename(name):
    ''' Make sure log filenames look like logs. '''
    if name.endswith('.log'):
        log_name = name
    else:
        log_name = f'{name}.log'
    return log_name

def set_alert_addresses(from_address, to_address):
    '''Set the mail addresses to use if serious error detected while logging.
    '''
    global alert_from_address, alert_to_address

    alert_from_address = from_address
    alert_to_address = to_address

def format_args(*args, separator=None):
    ''' Return args as text joined by separator.

        This is an alternative to the standard log args formatting of (roughly)::
            self.debug(message, *args)
                ...
                self.log(message.format(*args)(

        Separator defaults to ': '.
    '''

    def text(arg):
        ''' Filter arg to str. '''

        if isinstance(arg, Exception):
            # assume current exception
            result = format_exc()
        else:
            result = str(arg)

        return result

    separator = separator or ': '

    # remove empty args
    args = filter(bool, args)
    # convert to text
    args = list(map(text, args))

    line = separator.join(args)
    return line

def clear_user_logs(user):
    ''' Clear all logs created by this module for the user.

        The caller must have the correct file permissions.
    '''

    logdir = os.path.join(BASE_LOG_DIR, user)
    if os.listdir(logdir):
        run('rm', f'{logdir}/*')

    # if denova.python.log.DEBUGGING is True, then denova.python.log creates alt logs
    alt_log = f'/tmp/_log.{user}.log'
    if os.path.exists(alt_log):
        os.remove(alt_log)

    default_log = f'/tmp/python.default.{user}.log'
    if os.path.exists(default_log):
        os.remove(default_log)

def _test():
    """
        >>> import denova.python.log

        >>> # if we don't specify a log filename,
        >>> # then denova.python.log uses the calling module,
        >>> # in this case '<doctest...'
        >>> log = denova.python.log.get('_test.log')

        >>> log.filename
        '_test.log'

        >>> unique_message = 'test message to {} at {}'.format(log.filename, denova.python._log.timestamp())
        >>> log(unique_message)

        # give logwriter time to write
        >>> import time
        >>> time.sleep(1)

        >>> logdir = '/var/local/log/{}'.format(denova.python._log.whoami())
        >>> for basename in ['_test.log', 'master.log']:
        ...     logpath = os.path.join(logdir, basename)
        ...     with open(logpath) as logfile:
        ...         content = logfile.read()
        ...         assert unique_message in content
    """

    # there must be some body to the function other than comments
    # although prospector complains
    pass


# alternative name for get()
# for security it's best to import only what you need
# the code is clearer with a meaningful class or function name
# example:
#     from denova.python.log import get_log()
#     log = get_log()
get_log = get


if __name__ == "__main__":
    import doctest
    doctest.testmod()



#! /bin/env python3
'''
    Drive lib

    Copyright 2018-2020 DeNova
    Last modified: 2020-10-20
'''

import sys
import time
from subprocess import CalledProcessError

from denova.os.command import run

def list_block_devices():
    ''' List block devices, i.e. drives and partitions

        >>> blocks = list_block_devices()
    '''

    blocks = []
    lsblk_out = run('lsblk', '--list').stdout
    """ Example output::
            NAME MAJ:MIN RM   SIZE RO TYPE MOUNTPOINT
            sda    8:0    0 698.7G  0 disk
            sda1   8:1    0 694.8G  0 part /
            sda2   8:2    0     1K  0 part
            sda5   8:5    0   3.9G  0 part [SWAP]
            sdb    8:16   1 970.5M  0 disk
            sdb1   8:17   1   958M  0 part
            sr0   11:0    1  1024M  0 rom
    """
    lines = lsblk_out.split('\n')
    for line in lines:
        if line.startswith('s'):
            name = line.split()[0]
            blocks.append(name)

    return blocks

def list_drives():
    ''' List drives

        >>> drives = list_drives()
    '''

    drives = []
    for name in list_block_devices():
        # partitions end in a nonzero digit
        # '0' not included because 'sr0' is a drive
        drives.append(name.rstrip('123456789'))

    return sorted(set(drives))

def size(drive):
    ''' Return drive size in bytes. '''

    drive = full_dev_name(drive)
    size_str = run('blockdev', '--getsize64', drive).stdout
    return int(size_str)

def full_dev_name(drive):
    ''' Return full device name. E.g. /dev/sda

        If drive already starts with /dev/, it is returned unchanged.
    '''

    if not drive.startswith('/dev/'):
        drive = '/dev/' + drive
    return drive

def physically_select_drive():
    ''' Interactively require user to plug in the selected drive.

        Return device name of selected drive.

        Don't ask for the drive device name
        Have them safely specify the drive by plugging it in when requested

        Because this is a security measure, exits program on error.
    '''

    def diff_drives(drives_before):
        drives_now = list_drives()
        # debug('drives: {}'.format(drives_now)) # DEBUG
        new_drives = set(drives_now) - set(drives_before)
        # debug('new drives: {}'.format(new_drives)) # DEBUG
        return new_drives

    def get_new_drives():
        wait = 10 # seconds
        print("Be sure which drive you want to select.")
        print('')

        print('If the drive you are selecting is mounted, unmount it.')
        print('If the drive you are selecting is plugged in, unplug it.')
        print('')
        input('When the drive you are selecting is unmounted and unplugged, then press Enter.')
        drives_before = list_drives()

        print('')
        input('Plug in the drive you are selecting, then press Enter again.')

        # wait for the system to recognize the drive
        new_drives = diff_drives(drives_before)
        while (not new_drives) and wait:
            time.sleep(1)
            wait = wait - 1
            new_drives = diff_drives(drives_before)

        return new_drives

    try:
        new_drives = get_new_drives()

        if len(new_drives) > 1:
            sys.exit(f'more than one new drive: {new_drives}')

        elif len(new_drives) == 0:
            sys.exit('no new drive')

    except KeyboardInterrupt:
        # don't show traceback
        sys.exit('')

    else:
        drive = new_drives.pop()

    return full_dev_name(drive)

def uuid(device):
    ''' Return UUID for device, or if none return None. '''

    device = full_dev_name(device)

    # skip e.g. /dev/sr0
    if device.startswith('/dev/sd'):
        result = run('blkid', '-s', 'UUID', '-o', 'value', device).stdout
    else:
        result = None

    return result

def set_uuid(device, uuid):
    ''' Set UUID for device. '''

    run('tune2fs', '-U', uuid, full_dev_name(device))

def make_uuid():
    ''' Return a new UUID. '''

    return run('uuidgen').stdout

def deduplicate_uuids():
    ''' Interactively set new uuids for duplicate uuid drives .

        Some disk copy utilities duplicate uuids. Linux doesn't like that.
    '''

    def get_uuids():
        uuids = {}
        for block_device in list_block_devices():
            u = uuid(block_device)
            if u:
                uuids[full_dev_name(block_device)] = u
        return uuids

    print('Make sure all drives that may have duplicate uuids are plugged in.')
    input('Then press Enter')
    print('Waiting for system to recognize all drives. . .')
    time.sleep(10)
    print('')

    uuids = get_uuids()
    #print(uuids) # DEBUG
    unique_uuids = sorted(set(uuids.values()))

    # while there are dups
    while sorted(uuids.values()) != unique_uuids:

        for unique_uuid in unique_uuids:
            # get drives that match this uuid
            matches = []
            for block_device, u in uuids.items():
                if u == unique_uuid:
                    matches.append(block_device)

            # if this uuid has dups
            if len(matches) > 1:
                dup_block_devices = matches

                print(f"Found duplicate uuids for {','.join(dup_block_devices)}")
                print('Follow these instructions to physically select a drive, and change its uuid.')

                drive_to_change = physically_select_drive()

                # verify drive
                # partitions end in a digit, but not drives
                # this isn't quite right if multiple partitions
                # on a drive have the same uuid
                dup_drives = []
                for dup in dup_block_devices:
                    drive = dup.rstrip('0123456789')
                    if drive not in dup_drives:
                        dup_drives.append(drive)
                dup_drives = sorted(dup_drives)
                # print(dup_drives) # DEBUG
                while drive_to_change not in dup_drives:
                    print('')
                    print('{} is not in the list of duplicates: {}'.
                          format(drive_to_change, ','.join(dup_drives)))
                    drive_to_change = physically_select_drive()

                # find the matching block device, which may be a partition
                selected_block_device = None
                for block in dup_block_devices:
                    if block.startswith(drive_to_change):
                        if selected_block_device is None:
                            selected_block_device = block
                        else:
                            sys.exit(f'more than one matching block device on {drive_to_change}')
                if selected_block_device is None:
                    sys.exit(f'no matching block device on {drive_to_change}')
                #print(selected_block_device) # DEBUG

                print('')
                yesno = input('Are you sure you want to change the uuid for {}? (y/n)'.
                              format(selected_block_device))
                if yesno == 'y' or yesno == 'Y':
                    new_uuid = make_uuid()
                    try:
                        set_uuid(selected_block_device, new_uuid)
                    except CalledProcessError as cpe:
                        print(cpe.stdout)
                        print(cpe.stderr)
                    else:
                        print(f'Set uuid of {selected_block_device} to {new_uuid}')

        uuids = get_uuids()
        unique_uuids = sorted(set(uuids.values()))


if __name__ == "__main__":


    import doctest
    doctest.testmod()




'''
    Shared django settings for between projects.

    Copyright 2009-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

try:
    # only importing to verify django's installed
    import django
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

import os
from tempfile import gettempdir

from denova.os.user import whoami

DJANGO_ADDONS_PROJECT_PATH = os.path.dirname(__file__).replace('\\','/')

# probably want to include this directory in the TEMPLATES_DIR so the shared admin templates are used
DJANGO_ADDONS_TEMPLATE_DIR = os.path.join(DJANGO_ADDONS_PROJECT_PATH, 'templates')
DJANGO_ADDONS_STATIC_DIR = os.path.join(DJANGO_ADDONS_PROJECT_PATH, 'static')

APPEND_SLASH = True

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'Atlantic/Reykjavik'

# If you set this to True, Django will use timezone-aware datetimes.
# Our CRM requires this to be set True; adjust in app's settings if you want to override.
USE_TZ = True

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-CA'
LANGUAGES = [
    ('en', 'English'),
    ('en-CA', 'English-Canadian'),
]

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

USE_L10N = True

# reduce "database is locked" errors, especially with sqlite
DATABASE_OPTIONS = {'timeout': 30}

# django staticfiles app url prefix for static files
STATIC_URL = "/static/"

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

# login urls
LOGIN_REDIRECT_URL = '/'

# do not allow any pages from our site to be wrapped in frames
X_FRAME_OPTIONS = 'DENY'

INSTALLED_APPS = (
    'django.contrib.humanize',
    'django.contrib.staticfiles',

    # django third party
    #'debug_toolbar',
    'django_singleton_admin',

    # https://pypi.python.org/pypi/django-bootstrap4
    'bootstrap4',
    # https://pypi.python.org/pypi/django-forms-bootstrap/
    'django_forms_bootstrap',

    'django.contrib.admin',
    'django.contrib.auth',
)

# List of template loaders that know how to import templates from various sources.
TEMPLATE_LOADERS = [
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader'
]

# List of processors
TEMPLATE_CONTEXT_PROCESSORS = [
    'django.template.context_processors.debug',
    'django.template.context_processors.request',
    'django.contrib.auth.context_processors.auth',
    'django.contrib.messages.context_processors.messages',

    'denova.django_addons.context_processors.custom',
]

# order recommended by django 2.0 docs
MIDDLEWARE = (
    #'django.middleware.cache.UpdateCacheMiddleware',    # This must be first on the list to use caching
    'django.middleware.security.SecurityMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.BrokenLinkEmailsMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.locale.LocaleMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',

    'denova.django_addons.middleware.template.RequestMiddleware',
    'denova.django_addons.middleware.template.SettingsMiddleware',
    'denova.django_addons.middleware.debug.DebugMiddleware',
    #'denova.django_addons.middleware.debug.StripCodeComments',

    #'django.middleware.cache.FetchFromCacheMiddleware', # This must be last to use caching
)


# Log everything to a file. Log errors, such as HTTP 500 error when DEBUG=False, to email.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
DJANGO_USER = whoami()
DJANGO_LOG_DIR = f'/var/local/log/{DJANGO_USER}'
DJANGO_LOG_FILENAME = f'{DJANGO_LOG_DIR}/django.log'
try:
    if not os.path.exists(DJANGO_LOG_DIR):
        os.makedirs(DJANGO_LOG_DIR)
except Exception:
    try:
        # each log needs a unique name per user or there are permission conflicts
        DJANGO_LOG_FILENAME = os.path.join(gettempdir(), f'django_{DJANGO_USER}.log')
    except Exception:
        from tempfile import NamedTemporaryFile
        DJANGO_LOG_FILENAME = NamedTemporaryFile(mode='a', prefix='django', dir=gettempdir())

LOGGING_CONFIG = None
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'formatters': {
        'verbose': {
            'format': '%(asctime)s %(levelname)s %(module)s %(process)d %(thread)d %(message)s'
        },
        'simple': {
            'format': '[%(asctime)s %(module)s] %(levelname)s: %(message)s'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler',
            'formatter': 'verbose'
        },
        'file': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'formatter': 'verbose',
            'filename': DJANGO_LOG_FILENAME,
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['file', 'mail_admins'],
            'level': 'INFO',
            'propagate': True,
        },
    }
}



'''
    OS identification.

    Copyright 2008-2020 DeNova
    Last modified: 2020-09-30

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''

from platform import system

def get_os_name():
    '''
        Return the name of the OS in lower case.

        >>> get_os_name()
        'linux'
    '''

    return system().lower()

def is_windows():
    '''
        Return True if the underlying OS is any variant of Windows.

        >>> is_windows()
        False
    '''

    osname = get_os_name()
    return osname.startswith("win") or osname.startswith("microsoft windows")

def is_windows_vista():
    ''' Return True if the underlying OS is Windows Vista.

        >>> is_windows_vista()
        False
    '''
    winVista = False
    osname = get_os_name()
    if is_windows():
        winVista = osname.find("vista") > 0
    return winVista

def is_windows7():
    ''' Return True if the underlying OS is Windows 7.

        >>> is_windows7()
        False
    '''
    win7 = False
    osname = get_os_name()
    if is_windows():
        win7 = osname.find("7") > 0
    return win7

def is_windows8():
    ''' Return True if the underlying OS is Windows 8.

        >>> is_windows8()
        False
    '''
    win8 = False
    osname = get_os_name()
    if is_windows():
        win8 = osname.find("8") > 0
    return win8

def is_windows10():
    ''' Return True if the underlying OS is Windows 10.

        >>> is_windows10()
        False
    '''
    win10 = False
    osname = get_os_name()
    if is_windows():
        win10 = osname.find("10") > 0
    return win10

def is_windows_xp():
    ''' Return True if the underlying OS is Windows XP.

        >>> is_windows_xp()
        False
    '''
    winXP = False
    osname = get_os_name()
    if is_windows():
        #  sometimes XP falsely reports that its W2K
        if osname.find("xp") > 0 or osname.find("2000") > 0:
            winXP = True
    return winXP

def is_unix():
    ''' Return True if the underlying OS is any variant of unix.

        >>> is_unix()
        True
    '''
    osname = get_os_name()
    return osname.find("unix") >= 0 or is_linux() or is_aix() or is_hp_unix() or is_solaris() or is_mac_os_x()

def is_linux():
    ''' Return True if the underlying OS is Linux.

        >>> is_linux()
        True
    '''
    osname = get_os_name()
    return osname.find("linux") >= 0

def is_aix():
    ''' Return True if the underlying OS is IBM's AIX.

        >>> is_aix()
        False
    '''
    osname = get_os_name()
    return osname.find("aix") >= 0

def is_hp_unix():
    ''' Return True if the underlying OS is HP's unix.

        >>> is_hp_unix()
        False
    '''
    osname = get_os_name()
    return osname.find("hp-ux") >= 0 or osname.find("hpux") >= 0 or osname.find("irix") >= 0

def is_solaris():
    ''' Return True if the underlying OS is Solaris.

        >>> is_solaris()
        False
    '''
    osname = get_os_name()
    return osname.find("solaris") >= 0 or osname.find("sunos") >= 0

def is_mac_os_x():
    ''' Return True if the underlying OS is Mac OS X.

        >>> is_mac_os_x()
        False
    '''
    osname = get_os_name()
    return osname.find("mac os x") >= 0

def is_mac():
    ''' Return True if the underlying OS is any variant of Mac.

        >>> is_mac()
        False
    '''
    osname = get_os_name()
    return osname.find("mac os") >= 0 or osname.find("macos") >= 0




'''
    Get and save a singleton record.

    Copyright 2015-2020 DeNova
    Last modified: 2020-10-20

    This file is open source, licensed under GPLv3 <http://www.gnu.org/licenses/>.
'''
from traceback import format_exc

try:
    from django.db import transaction
except ModuleNotFoundError:
    import sys
    sys.exit('Django required')

from denova.python.log import get_log

log = get_log()

def get_singleton(model, db=None):
    '''
        Get a singleton record.

        >>> from django.contrib.auth.models import Group
        >>> records = Group.objects.all()
        >>> for record in records:
        ...     __ = record.delete()
        >>> record = Group(name='special')
        >>> save_singleton(Group, record)
        >>> singleton_record = get_singleton(Group)
        >>> record == singleton_record
        True
    '''

    record = None

    with transaction.atomic():
        if db is None:
            records = model.objects.all()
        else:
            records = model.objects.using(db).all()

        if records.count() == 1:
            for r in records:
                record = r
        elif records.count() > 1:
            for r in records:
                if record is None:
                    record = r
                else:
                    log(f'deleted extra record: {r.pk}')
                    r.delete()

        if record is None:
            # the higher level must handle this case
            raise model.DoesNotExist()

    return record


def save_singleton(model, record, db=None):
    '''
        Save a singleton record.

        >>> from django.contrib.auth.models import Group
        >>> records = Group.objects.all()
        >>> for record in records:
        ...     __ = record.delete()
        >>> record = Group(name='special')
        >>> save_singleton(Group, record)
        >>> singleton_record = get_singleton(Group)
        >>> record == singleton_record
        True
    '''

    try:
        with transaction.atomic():
            if db is None:
                record.save()
            else:
                record.save(using=db)

        # get the singleton again to insure there's only 1 record
        get_singleton(model, db=db)
    except Exception:
        log(f'tried to save {model}')
        log(format_exc())
        raise
